#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.symbols as commons_symbols

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data
from tests.test_utils.random_numbers import decimal_random_quantity, decimal_random_price

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_get_current_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.get_current_crypto_currencies_values() == \
           {'BTC': constants.ONE, 'USDT': constants.ZERO}
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'ETH': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'XRP': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'DOT': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'MATIC': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()},
        'USDT': {'available': decimal_random_quantity(), 'total': decimal_random_quantity()}
    }, True)
    portfolio_manager.handle_balance_updated()

    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': constants.ZERO,
        'XRP': constants.ZERO,
        'DOT': constants.ZERO,
        'MATIC': constants.ZERO,
        'USDT': constants.ZERO
    }

    exchange_manager.client_symbols.append("MATIC/BTC")
    exchange_manager.client_symbols.append("XRP/BTC")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("XRP")
    portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal("0.005"))
    exchange_manager.client_symbols.append("DOT/BTC")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("DOT")
    portfolio_manager.handle_mark_price_update("DOT/BTC", decimal.Decimal("0.05"))
    exchange_manager.client_symbols.append("BTC/USDT")

    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': constants.ZERO,
        'XRP': decimal.Decimal("0.005"),
        'DOT': decimal.Decimal("0.05"),
        'MATIC': constants.ZERO,
        'USDT': constants.ZERO
    }
    matic_btc_price = decimal_random_price(max_value=decimal.Decimal(0.05))
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("MATIC")
    portfolio_manager.handle_mark_price_update("MATIC/BTC", matic_btc_price)
    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': constants.ZERO,
        'XRP': decimal.Decimal("0.005"),
        'DOT': decimal.Decimal("0.05"),
        'MATIC': matic_btc_price,
        'USDT': constants.ZERO
    }
    usdt_btc_price = decimal_random_price(max_value=decimal.Decimal('0.01'))
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("USDT")
    portfolio_manager.handle_mark_price_update("BTC/USDT", usdt_btc_price)
    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': constants.ZERO,
        'XRP': decimal.Decimal("0.005"),
        'DOT': decimal.Decimal("0.05"),
        'MATIC': matic_btc_price,
        'USDT': constants.ONE / usdt_btc_price
    }
    eth_btc_price = decimal_random_price(max_value=constants.ONE)
    exchange_manager.client_symbols.append("ETH/BTC")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("ETH")
    portfolio_manager.handle_mark_price_update("ETH/BTC", eth_btc_price)
    assert portfolio_value_holder.get_current_crypto_currencies_values() == {
        'BTC': constants.ONE,
        'ETH': decimal.Decimal(str(eth_btc_price)),
        'XRP': decimal.Decimal("0.005"),
        'DOT': decimal.Decimal("0.05"),
        'MATIC': matic_btc_price,
        'USDT': constants.ONE / usdt_btc_price
    }


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_get_current_holdings_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    exchange_manager.client_symbols.append("ETH/BTC")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.discard("ETH")
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal.Decimal("10"), 'total': decimal.Decimal("10")},
        'ETH': {'available': decimal.Decimal("100"), 'total': decimal.Decimal("100")},
        'XRP': {'available': decimal.Decimal("10000"), 'total': decimal.Decimal("10000")},
        'USDT': {'available': decimal.Decimal("1000"), 'total': decimal.Decimal("1000")}
    }, True)
    portfolio_manager.handle_balance_updated()
    assert portfolio_value_holder.get_current_holdings_values() == {
        'BTC': decimal.Decimal("10"),
        'ETH': constants.ZERO,
        'XRP': constants.ZERO,
        'USDT': constants.ZERO
    }
    def mock_get_symbol_position(symbol, side):
        mock_pos = mock.Mock()
        if symbol == "ETH/USDT":
            mock_pos.margin = decimal.Decimal("5000")
            mock_pos.is_idle.return_value = False
        elif symbol == "XRP/USDT":
            mock_pos.margin = decimal.Decimal("0.1")
            mock_pos.is_idle.return_value = False
        else:
            raise errors.ContractExistsError(f"Contract {symbol} does not exist")
        return mock_pos

    def mock_create_symbol_position(symbol, position_id):
        mock_pos = mock.Mock()
        mock_pos.position_id = position_id
        mock_pos.symbol = symbol
        mock_pos.margin = constants.ZERO
        mock_pos.is_idle.return_value = True
        return mock_pos

    with mock.patch.object(
        exchange_manager.exchange_personal_data.positions_manager,
        "_create_symbol_position",
        mock_create_symbol_position,
    ), mock.patch.object(
        exchange_manager.exchange_personal_data.positions_manager,
        "get_symbol_position",
        mock_get_symbol_position,
    ):
        portfolio_value_holder.value_converter.last_prices_by_trading_pair["ETH/BTC:BTC"] = decimal.Decimal("50")
        portfolio_value_holder.value_converter.last_prices_by_trading_pair["ETH/BTC"] = decimal.Decimal("50")
        # Update current_crypto_currencies_values to include ETH with the calculated price
        portfolio_value_holder.current_crypto_currencies_values["ETH"] = decimal.Decimal("50")
        portfolio_value_holder.sync_portfolio_current_value_using_available_currencies_values(init_price_fetchers=False)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal("10"),
            'ETH': decimal.Decimal("5000"),
            'XRP': constants.ZERO,
            'USDT': constants.ZERO
        }
        portfolio_manager.handle_mark_price_update("XRP/USDT", 2.5)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal("10"),
            'ETH': decimal.Decimal("5000"),
            'XRP': constants.ZERO,
            'USDT': constants.ZERO
        }
        portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal('0.00001'))
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal("10"),
            'ETH': decimal.Decimal("5000"),
            'XRP': constants.ZERO,
            'USDT': constants.ZERO
        }
        exchange_manager.client_symbols.append("XRP/BTC")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("XRP")
        portfolio_manager.handle_mark_price_update("XRP/BTC", decimal.Decimal('0.00001'))
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal(10),
            'ETH': decimal.Decimal(5000),
            'XRP': decimal.Decimal(str(0.1)),
            'USDT': constants.ZERO
        }
        exchange_manager.client_symbols.append("BTC/USDT")
        portfolio_value_holder.value_converter.missing_currency_data_in_exchange.remove("USDT")
        portfolio_manager.handle_mark_price_update("BTC/USDT", 5000)
        assert portfolio_value_holder.get_current_holdings_values() == {
            'BTC': decimal.Decimal(10),
            'ETH': decimal.Decimal(5000),
            'XRP': decimal.Decimal(str(0.1)),
            'USDT': decimal.Decimal(str(0.2))
        }


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_get_origin_portfolio_current_value(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == decimal.Decimal(str(10))

@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_get_origin_portfolio_current_value_with_different_reference_market(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    portfolio_manager.reference_market = "USDT"
    portfolio_manager.handle_profitability_recalculation(True)
    assert portfolio_value_holder.get_origin_portfolio_current_value() == decimal.Decimal(str(1000))

@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_update_origin_crypto_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.update_origin_crypto_currencies_values("ETH/BTC", decimal.Decimal(str(0.1))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["ETH"] == decimal.Decimal(str(0.1))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["ETH/BTC"] == decimal.Decimal(str(0.1))
    # ETH is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("ETH/BTC", decimal.Decimal(str(0.1))) is False

    assert portfolio_value_holder.update_origin_crypto_currencies_values("BTC/USDT", decimal.Decimal(str(100))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["USDT"] == \
           decimal.Decimal(constants.ONE / decimal.Decimal(100))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["BTC/USDT"] == decimal.Decimal(str(100))
    # USDT is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("BTC/USDT", decimal.Decimal(str(100))) is False

    # with bridge pair (DOT/ETH -> ETH/BTC to compute DOT/BTC)
    assert portfolio_value_holder.update_origin_crypto_currencies_values("DOT/ETH", decimal.Decimal(str(0.015))) is True
    assert portfolio_value_holder.origin_crypto_currencies_values["DOT"] == \
           decimal.Decimal(str(0.015)) * decimal.Decimal(str(0.1))
    assert portfolio_value_holder.value_converter.last_prices_by_trading_pair["DOT/ETH"] == decimal.Decimal(str(0.015))
    # USDT is now priced and BTC is the reference market
    assert portfolio_value_holder.update_origin_crypto_currencies_values("DOT/ETH", decimal.Decimal(str(0.015))) \
           is False

@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_sync_portfolio_current_value_using_available_currencies_values(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder

    assert portfolio_value_holder.portfolio_current_value == constants.ZERO
    portfolio_value_holder.sync_portfolio_current_value_using_available_currencies_values()
    assert portfolio_value_holder.portfolio_current_value == decimal.Decimal(str(10))

    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.clear()
    exchange_manager.client_symbols.append("BTC/USDT")
    portfolio_manager.handle_mark_price_update("BTC/USDT", decimal.Decimal(str(100)))
    portfolio_value_holder.sync_portfolio_current_value_using_available_currencies_values()
    assert portfolio_value_holder.portfolio_current_value == decimal.Decimal(str(20)) # now includes USDT

@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "futures"], indirect=True)
async def test_get_holdings_ratio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    symbol = "BTC/USDT"
    exchange_manager.client_symbols = [symbol]
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair[symbol] = decimal.Decimal("1000")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        portfolio_current_value = decimal.Decimal("11")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio = {}
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["BTC"] = \
        personal_data.SpotAsset(name="BTC", available=decimal.Decimal("10"), total=decimal.Decimal("10"))
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["USDT"] = \
        personal_data.SpotAsset(name="USDT", available=decimal.Decimal("1000"), total=decimal.Decimal("1000"))
    usdt_removed = [False]

    def mock_get_or_create_position_impl(symbol, side):
        mock_pos = mock.Mock()
        if symbol == "BTC/USDT:USDT":
            if usdt_removed[0]:
                mock_pos.margin = constants.ZERO
                mock_pos.is_idle.return_value = True
            else:
                mock_pos.margin = decimal.Decimal("1")
                mock_pos.is_idle.return_value = False
        elif symbol == "BTC/XYZ:XYZ":
            mock_pos.margin = decimal.Decimal("0")
            mock_pos.is_idle.return_value = False
        else:
            raise errors.ContractExistsError(f"Contract {symbol} does not exist")
        return mock_pos

    mock_get_or_create = mock.Mock(side_effect=mock_get_or_create_position_impl)

    def mock_get_contract(pair):
        mock_contract = mock.Mock()
        mock_contract.is_one_way_position_mode.return_value = True
        return mock_contract

    with mock.patch.object(
        exchange_manager.exchange_personal_data.positions_manager,
        "_get_or_create_position",
        mock_get_or_create,
    ), mock.patch.object(
        exchange_manager.exchange,
        "get_pair_future_contract", # TODO update this mock to use get_pair_contract to support options
        mock_get_contract,
    ):
        exchange_manager.exchange_config.traded_symbols = [commons_symbols.parse_symbol("BTC/USDT")]
        assert portfolio_value_holder.get_holdings_ratio("BTC") == decimal.Decimal('0.9090909090909090909090909091')
        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=False) \
               == decimal.Decimal('0.9090909090909090909090909091')
        assert portfolio_value_holder.get_holdings_ratio("USDT") == decimal.Decimal('0.09090909090909090909090909091')

        exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio.pop("USDT")
        usdt_removed[0] = True
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
            portfolio_current_value = decimal.Decimal("10")
        assert portfolio_value_holder.get_holdings_ratio("BTC") == constants.ONE
        # add ETH and try to get ratio without symbol price
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio.\
            get_currency_portfolio("ETH").total = decimal.Decimal(10)
        # force not backtesting mode
        exchange_manager.is_backtesting = False
        # force add symbol in exchange symbols
        exchange_manager.client_symbols.append("ETH/BTC")
        if not exchange_manager.is_spot_only:
            with pytest.raises(errors.ContractExistsError):
                _ = portfolio_value_holder.get_holdings_ratio("ETH")
        else:
            with pytest.raises(errors.MissingPriceDataError):
                _ = portfolio_value_holder.get_holdings_ratio("ETH")
        # let channel register proceed
        await asyncio_tools.wait_asyncio_next_cycle()
        assert portfolio_value_holder.get_holdings_ratio("BTC") == constants.ONE
        assert portfolio_value_holder.get_holdings_ratio("USDT") == constants.ZERO
        assert portfolio_value_holder.get_holdings_ratio("XYZ") == constants.ZERO

        # without traded_symbols
        exchange_manager.exchange_config.traded_symbols.clear()
        assert portfolio_value_holder.get_holdings_ratio("BTC", traded_symbols_only=True) == constants.ZERO

        # with traded_symbols
        exchange_manager.exchange_config.traded_symbols.extend([
            commons_symbols.parse_symbol("BTC/USDT"), commons_symbols.parse_symbol("ETH/USDT")
        ])
        assert portfolio_value_holder.get_holdings_ratio("BTC", traded_symbols_only=True) == constants.ONE
 
@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
async def test_get_holdings_ratio_with_include_assets_in_open_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    symbol = "BTC/USDT"
    has_eth_position = False
    exchange_manager.client_symbols = [symbol]
    exchange_manager.exchange_config.traded_symbols = [commons_symbols.parse_symbol("BTC/USDT")]
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair[symbol] = decimal.Decimal("1000")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair["BTC/USDT:BTC"] = decimal.Decimal("1000")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair["BTC/USDT:USDT"] = decimal.Decimal("1000")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair["ETH/USDT"] = decimal.Decimal("100")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair["ETH/USDT:BTC"] = decimal.Decimal("100")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        value_converter.last_prices_by_trading_pair["ETH/BTC:BTC"] = decimal.Decimal("0.1")
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        portfolio_current_value = decimal.Decimal("11")
    exchange_manager.exchange_personal_data.portfolio_manager.reference_market = "BTC"
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio = {}
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["BTC"] = \
        personal_data.SpotAsset(name="BTC", available=decimal.Decimal("10"), total=decimal.Decimal("10"))
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["USDT"] = \
        personal_data.SpotAsset(name="USDT", available=decimal.Decimal("1000"), total=decimal.Decimal("1000"))
    def mock_get_or_create_position_impl(symbol, side):
        mock_pos = mock.Mock()
        if symbol == "BTC/USDT:BTC":
            mock_pos.margin = decimal.Decimal("10") # equivalent to the 10 BTC in the spot portfolio
            mock_pos.is_idle.return_value = False
        if symbol == "BTC/USDT:USDT":
            mock_pos.margin = decimal.Decimal("1") # equivalent to the 1000 USDT vs BTC in the spot portfolio
            mock_pos.is_idle.return_value = False
        elif symbol == "ETH/USDT:BTC":
            mock_pos.margin = decimal.Decimal("100") if has_eth_position else constants.ZERO # equivalent to the 10 ETH vs BTC in the spot portfolio
            mock_pos.is_idle.return_value = False
        elif symbol == "ETH/BTC:BTC":
            mock_pos.margin = decimal.Decimal("1") if has_eth_position else constants.ZERO # equivalent to the 10 ETH vs BTC in the spot portfolio
            mock_pos.is_idle.return_value = False
        else:
            raise errors.ContractExistsError(f"Contract {symbol} does not exist")
        return mock_pos

    mock_get_or_create = mock.Mock(side_effect=mock_get_or_create_position_impl)

    def mock_get_contract(pair):
        mock_contract = mock.Mock()
        mock_contract.is_one_way_position_mode.return_value = True
        return mock_contract

    with mock.patch.object(
        exchange_manager.exchange_personal_data.positions_manager,
        "_get_or_create_position",
        mock_get_or_create,
    ), mock.patch.object(
        exchange_manager.exchange,
        "get_pair_future_contract",
        mock_get_contract,
    ):
        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=False) \
               == decimal.Decimal('0.9090909090909090909090909091')
        assert portfolio_value_holder.get_holdings_ratio("USDT", include_assets_in_open_orders=False) \
               == decimal.Decimal('0.09090909090909090909090909091')

        # no open order
        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.9090909090909090909090909091')
        assert portfolio_value_holder.get_holdings_ratio("USDT", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.09090909090909090909090909091')

        order_1 = mock.Mock(
            order_id="1", symbol="BTC/USDT", origin_quantity=decimal.Decimal("1"), total_cost=decimal.Decimal('1000'),
            status=enums.OrderStatus.OPEN, side=enums.TradeOrderSide.BUY
        )
        order_2 = mock.Mock(
            order_id="2", symbol="ETH/USDT", origin_quantity=decimal.Decimal("2"), total_cost=decimal.Decimal('100'),
            status=enums.OrderStatus.OPEN, side=enums.TradeOrderSide.BUY
        )
        order_3 = mock.Mock(
            order_id="3", symbol="XRP/ETH", origin_quantity=decimal.Decimal("100"), total_cost=decimal.Decimal('0.001'),
            status=enums.OrderStatus.OPEN, side=enums.TradeOrderSide.SELL
        )
        # open orders
        await exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order_1)
        await exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order_2)
        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=False) \
               == decimal.Decimal('0.9090909090909090909090909091')
        assert portfolio_value_holder.get_holdings_ratio("ETH", include_assets_in_open_orders=False) \
               == decimal.Decimal('0')
        assert portfolio_value_holder.get_holdings_ratio("USDT", include_assets_in_open_orders=False) \
               == decimal.Decimal('0.09090909090909090909090909091')

        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=True) \
               == decimal.Decimal('1')
        assert portfolio_value_holder.get_holdings_ratio("ETH", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.01818181818181818181818181818')
        assert portfolio_value_holder.get_holdings_ratio("USDT", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.09090909090909090909090909091')

        await exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order_3)
        assert portfolio_value_holder.get_holdings_ratio("BTC", include_assets_in_open_orders=True) \
               == decimal.Decimal('1')
        assert portfolio_value_holder.get_holdings_ratio("ETH", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.01819090909090909090909090909')
        assert portfolio_value_holder.get_holdings_ratio("USDT", include_assets_in_open_orders=True) \
               == decimal.Decimal('0.09090909090909090909090909091')

        # without traded_symbols
        exchange_manager.exchange_config.traded_symbols.clear()
        assert portfolio_value_holder.get_holdings_ratio(
            "BTC", traded_symbols_only=True, include_assets_in_open_orders=True
        ) == constants.ZERO

        # with traded_symbols
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["ETH"] = \
            personal_data.SpotAsset(name="ETH", available=decimal.Decimal("10"), total=decimal.Decimal("10"))
        has_eth_position = True
        exchange_manager.exchange_config.traded_symbols.extend([
            commons_symbols.parse_symbol("BTC/USDT")
        ])
        assert portfolio_value_holder.get_holdings_ratio(
            "BTC", traded_symbols_only=True, include_assets_in_open_orders=True
        ) == constants.ONE
        # 10 ETH (portfolio) + 2.001 (orders) in BTC / 11 = 0.1091
        assert portfolio_value_holder.get_holdings_ratio(
                "ETH", traded_symbols_only=True, include_assets_in_open_orders=True
        ) == decimal.Decimal('0.1091')
            
        # ETH now in traded assets
        exchange_manager.exchange_config.traded_symbols.extend([
            commons_symbols.parse_symbol("ETH/USDT")
        ])
        # 12.001 ETH in BTC / 12 total = 0.1000083
        assert portfolio_value_holder.get_holdings_ratio(
                "ETH", traded_symbols_only=True, include_assets_in_open_orders=True
        ) == decimal.Decimal('0.1000083333333333333333333333')  # ETH now taken into account in total value
            
        # ETH now taken into account in total value: BTC % of holdings is not 1 anymore (as ETH takes a part of this %)
        assert portfolio_value_holder.get_holdings_ratio(
            "BTC", traded_symbols_only=True, include_assets_in_open_orders=True
        ) == decimal.Decimal('0.9166666666666666666666666667')


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
@pytest.mark.parametrize("unit,coins_whitelist", [
    ("BTC", None),
    ("BTC", ["BTC"]),
    ("BTC", ["USDT"]),
    ("BTC", ["BTC", "ETH"]),
])
async def test_get_traded_assets_holdings_value(backtesting_trader, unit, coins_whitelist):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal.Decimal("10"), 'total': decimal.Decimal("10")},
        'USDT': {'available': decimal.Decimal("1000"), 'total': decimal.Decimal("1000")},
        'ETH': {'available': decimal.Decimal("5"), 'total': decimal.Decimal("5")},
    }, True)
    
    exchange_manager.exchange_config.traded_symbols = [commons_symbols.parse_symbol("BTC/USDT")]
    exchange_manager.client_symbols.append("BTC/USDT")
    portfolio_value_holder.value_converter.last_prices_by_trading_pair["BTC/USDT"] = decimal.Decimal("1000")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.discard("USDT")
    
    result = portfolio_value_holder.get_traded_assets_holdings_value(unit, coins_whitelist)
    
    # Verify result is a decimal and non-negative
    assert isinstance(result, decimal.Decimal)
    assert result >= constants.ZERO


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
@pytest.mark.parametrize("currency,traded_symbols_only,include_assets_in_open_orders,coins_whitelist,expected_ratio_zero", [
    ("BTC", False, False, None, False),
    ("BTC", True, False, None, False),
    ("USDT", False, False, None, False),
    ("BTC", False, True, None, False),
    ("BTC", True, True, ["BTC"], False),
    ("XYZ", False, False, None, True),
])
async def test_get_holdings_ratio_from_portfolio(backtesting_trader, currency, traded_symbols_only, include_assets_in_open_orders, coins_whitelist, expected_ratio_zero):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal.Decimal("10"), 'total': decimal.Decimal("10")},
        'USDT': {'available': decimal.Decimal("1000"), 'total': decimal.Decimal("1000")},
    }, True)
    
    exchange_manager.exchange_config.traded_symbols = [commons_symbols.parse_symbol("BTC/USDT")]
    exchange_manager.client_symbols.append("BTC/USDT")
    portfolio_value_holder.value_converter.last_prices_by_trading_pair["BTC/USDT"] = decimal.Decimal("1000")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.discard("USDT")
    portfolio_value_holder.sync_portfolio_current_value_using_available_currencies_values(init_price_fetchers=False)
    
    result = portfolio_value_holder._get_holdings_ratio_from_portfolio(
        currency, traded_symbols_only, include_assets_in_open_orders, coins_whitelist
    )
    
    assert isinstance(result, decimal.Decimal)
    assert result >= constants.ZERO
    assert result <= constants.ONE
    if expected_ratio_zero:
        assert result == constants.ZERO


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
@pytest.mark.parametrize("coins_whitelist,traded_symbols_only,use_portfolio_current_value", [
    (None, False, True),
    (None, True, False),
    (["BTC", "USDT"], False, False),
    (["BTC"], True, False),
])
async def test_get_total_holdings_value(backtesting_trader, coins_whitelist, traded_symbols_only, use_portfolio_current_value):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    
    portfolio_manager.portfolio.update_portfolio_from_balance({
        'BTC': {'available': decimal.Decimal("10"), 'total': decimal.Decimal("10")},
        'USDT': {'available': decimal.Decimal("1000"), 'total': decimal.Decimal("1000")},
    }, True)
    
    exchange_manager.exchange_config.traded_symbols = [commons_symbols.parse_symbol("BTC/USDT")]
    exchange_manager.client_symbols.append("BTC/USDT")
    portfolio_value_holder.value_converter.last_prices_by_trading_pair["BTC/USDT"] = decimal.Decimal("1000")
    portfolio_value_holder.value_converter.missing_currency_data_in_exchange.discard("USDT")
    portfolio_value_holder.sync_portfolio_current_value_using_available_currencies_values(init_price_fetchers=False)
    
    result = portfolio_value_holder._get_total_holdings_value(coins_whitelist, traded_symbols_only)
    
    assert isinstance(result, decimal.Decimal)
    assert result >= constants.ZERO
    if use_portfolio_current_value and not coins_whitelist and not traded_symbols_only:
        assert result == portfolio_value_holder.portfolio_current_value


@pytest.mark.parametrize("backtesting_exchange_manager", ["spot", "margin", "futures", "options"], indirect=True)
@pytest.mark.parametrize("currency,order_symbol,order_side,order_quantity,order_cost,expected_delta", [
    ("BTC", "BTC/USDT", enums.TradeOrderSide.BUY, decimal.Decimal("1"), decimal.Decimal("1000"), decimal.Decimal("1")),
    ("USDT", "BTC/USDT", enums.TradeOrderSide.SELL, decimal.Decimal("1"), decimal.Decimal("1000"), decimal.Decimal("1000")),
    ("ETH", "ETH/BTC", enums.TradeOrderSide.BUY, decimal.Decimal("2"), decimal.Decimal("0.1"), decimal.Decimal("2")),
    ("BTC", "ETH/BTC", enums.TradeOrderSide.SELL, decimal.Decimal("2"), decimal.Decimal("0.1"), decimal.Decimal("0.1")),
    ("XYZ", "BTC/USDT", enums.TradeOrderSide.BUY, decimal.Decimal("1"), decimal.Decimal("1000"), constants.ZERO),
])
async def test_get_orders_delta(backtesting_trader, currency, order_symbol, order_side, order_quantity, order_cost, expected_delta):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    portfolio_value_holder = portfolio_manager.portfolio_value_holder
    
    mock_order = mock.Mock()
    mock_order.symbol = order_symbol
    mock_order.side = order_side
    mock_order.origin_quantity = order_quantity
    mock_order.total_cost = order_cost
    mock_order.status = enums.OrderStatus.OPEN
    
    with mock.patch.object(
        portfolio_manager.exchange_manager.exchange_personal_data.orders_manager,
        "get_open_orders",
        return_value=[mock_order]
    ):
        result = portfolio_value_holder._get_orders_delta(currency)
        
        assert isinstance(result, decimal.Decimal)
        assert result == expected_delta



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import asyncio

import octobot_trading.constants
import octobot_trading.enums as enums
import octobot_trading.errors
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.personal_data.orders.order_state as order_state


class PendingCreationOrderState(order_state.OrderState):

    def __init__(self, order, is_from_exchange_data, enable_associated_orders_creation=True,
        is_already_counted_in_available_funds=False
    ):
        super().__init__(
            order, is_from_exchange_data, enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
        self.state = enums.States.PENDING_CREATION

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask OrdersChannel Internal producer to refresh the order from the exchange
        :param force_synchronization: When True, for the update of the order from the exchange
        :return: the result of OrdersProducer.update_order_from_exchange()
        """
        try:
            t0 = time.time()
            iteration = 0
            # Retries might be necessary if the order is being filled while requesting and the 1st refresh
            # is not updating the state.
            # Loop here until we get a clear answer as the order is not yet open and therefore not in orders manager.
            # If it turns out to be instantly closed, OctoBot will miss it as is will never be fetched with
            # open orders refresh.
            self.ensure_not_cleared(self.order)
            self.get_logger().info(
                f"Synchronizing [{self._underlying_refreshed_state.value}] order {self.order} "
                f"with {self.order.exchange_manager.exchange_name} exchange"
            )
            while self.order.is_pending_creation() \
                    and time.time() - t0 < octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT:
                iteration += 1
                self.synchronization_attempts += 1
                await exchange_channel.get_chan(
                    octobot_trading.constants.ORDERS_CHANNEL,
                    self.order.exchange_manager.id
                ).get_internal_producer().update_order_from_exchange(
                    order=self.order,
                    wait_for_refresh=True,
                    force_job_execution=True,
                )
                if self.order.is_pending_creation():
                    message = f"Failed to receive an order update ({iteration} attempts). " \
                      f"Retrying in {self.PENDING_REFRESH_INTERVAL} seconds."
                    self.ensure_not_cleared(self.order)
                    if self.order.exchange_manager.exchange.EXPECT_POSSIBLE_ORDER_NOT_FOUND_DURING_ORDER_CREATION:
                        self.get_logger().debug(message)
                    else:
                        self.get_logger().error(
                            f"{message} This works but is unexpected on {self.order.exchange_manager.exchange_name}. "
                            f"Please report it if you see it."
                        )
                    await asyncio.sleep(self.PENDING_REFRESH_INTERVAL)
                    self.ensure_not_cleared(self.order)
            if self.order.is_pending_creation():
                self.get_logger().error(
                    f"Order state is still pending after {octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT}s and "
                    f"{iteration} retries. Something is wrong."
                )
            else:
                if iteration > 1:
                    self.get_logger().debug(f"Received pending order state update after {iteration} iterations")
        except octobot_trading.errors.InvalidOrderState:
            self.get_logger().debug(f"Skipping exchange synchronisation as order has already been closed.")

    def is_created(self) -> bool:
        """
        :return: True if the Order is created
        """
        # order is created on exchange but not open yet
        return True

    def is_open(self) -> bool:
        """
        :return: True if the Order is considered as open
        """
        return False

    async def on_refresh_successful(self):
        pass

    async def terminate(self):
        """
        Should wait for being replaced by an OpenOrderState or be cancelled
        """
        self.log_event_message(enums.StatesMessages.PENDING_CREATION)

    def is_pending(self) -> bool:
        return True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.enums as enums

import octobot_trading.constants as trading_constants
import octobot_trading.personal_data as trading_personal_data
from tests.personal_data import DEFAULT_MARKET_QUANTITY, DEFAULT_ORDER_SYMBOL
from tests.test_utils.random_numbers import decimal_random_quantity

from tests import event_loop
from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests.personal_data.orders import backtesting_buy_and_sell_limit_orders

pytestmark = pytest.mark.asyncio


async def test_update_price_if_outdated(backtesting_buy_and_sell_limit_orders):
    buy_limit_order, sell_limit_order = backtesting_buy_and_sell_limit_orders
    buy_order_price = decimal.Decimal("100")
    buy_limit_order.update(
        price=buy_order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / buy_order_price),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.BUY_LIMIT,
    )
    sell_order_price = decimal.Decimal("150")
    sell_limit_order.update(
        price=sell_order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / sell_order_price),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.BUY_LIMIT,
    )
    origin_update_limit_price_if_necessary = buy_limit_order._update_limit_price_if_necessary
    # buy side
    with mock.patch.object(
        buy_limit_order, "_update_limit_price_if_necessary",
        mock.Mock(side_effect=origin_update_limit_price_if_necessary)
    ) as _update_limit_price_if_necessary_mock:
        # without price
        await buy_limit_order.update_price_if_outdated()
        _update_limit_price_if_necessary_mock.assert_not_called()

        # with up-to-date price
        buy_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(buy_limit_order.symbol).\
            prices_manager.set_mark_price(decimal.Decimal("110"), enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)
        await buy_limit_order.update_price_if_outdated()
        _update_limit_price_if_necessary_mock.assert_called_once()
        assert buy_limit_order.origin_price is buy_order_price
        _update_limit_price_if_necessary_mock.reset_mock()

        # with outdated price
        buy_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(buy_limit_order.symbol).\
            prices_manager.set_mark_price(decimal.Decimal("90"), enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)
        await buy_limit_order.update_price_if_outdated()

        _update_limit_price_if_necessary_mock.assert_called_once()
        # price got adapted
        assert buy_limit_order.origin_price < buy_order_price
        assert buy_limit_order.origin_price == decimal.Decimal("90") * (
            trading_constants.ONE + trading_constants.CHAINED_ORDERS_OUTDATED_PRICE_ALLOWANCE
        )

    sell_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(sell_limit_order.symbol). \
        prices_manager._reset_prices()
    origin_update_limit_price_if_necessary = sell_limit_order._update_limit_price_if_necessary
    # sell side
    with mock.patch.object(
        sell_limit_order, "_update_limit_price_if_necessary",
        mock.Mock(side_effect=origin_update_limit_price_if_necessary)
    ) as _update_limit_price_if_necessary_mock:
        # without price
        await sell_limit_order.update_price_if_outdated()
        _update_limit_price_if_necessary_mock.assert_not_called()

        # with up-to-date price
        sell_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(sell_limit_order.symbol).\
            prices_manager.set_mark_price(decimal.Decimal("110"), enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)
        await sell_limit_order.update_price_if_outdated()
        _update_limit_price_if_necessary_mock.assert_called_once()
        assert sell_limit_order.origin_price is sell_order_price
        _update_limit_price_if_necessary_mock.reset_mock()

        # with outdated price
        not_round_price = decimal.Decimal("155") + decimal.Decimal(1/3)  # force not round number
        sell_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(sell_limit_order.symbol).\
            prices_manager.set_mark_price(not_round_price, enums.MarkPriceSources.EXCHANGE_MARK_PRICE.value)
        await sell_limit_order.update_price_if_outdated()

        _update_limit_price_if_necessary_mock.assert_called_once()
        # price got adapted
        assert sell_limit_order.origin_price > sell_order_price
        assert sell_limit_order.origin_price != not_round_price * (
            trading_constants.ONE - trading_constants.CHAINED_ORDERS_OUTDATED_PRICE_ALLOWANCE
        )
        # ensure price decimals got adapted
        assert sell_limit_order.origin_price == trading_personal_data.decimal_adapt_price(
            sell_limit_order.exchange_manager.exchange.get_market_status(sell_limit_order.symbol, with_fixer=False),
            not_round_price * (
                trading_constants.ONE - trading_constants.CHAINED_ORDERS_OUTDATED_PRICE_ALLOWANCE
            )
        )

async def test_filled_maker_or_taker(backtesting_buy_and_sell_limit_orders):
    buy_limit_order, sell_limit_order = backtesting_buy_and_sell_limit_orders
    buy_order_price = decimal.Decimal("100")
    buy_limit_order.update(
        price=buy_order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / buy_order_price),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.BUY_LIMIT,
    )
    assert buy_limit_order._filled_maker_or_taker() == enums.ExchangeConstantsMarketPropertyColumns.MAKER.value
    buy_limit_order.created_last_price = decimal.Decimal("99")
    assert buy_limit_order._filled_maker_or_taker() == enums.ExchangeConstantsMarketPropertyColumns.TAKER.value
    await buy_limit_order.set_as_chained_order(sell_limit_order, False, {}, False)
    sell_limit_order.origin_price = decimal.Decimal("101")
    # now uses trigger order filling price
    assert buy_limit_order._filled_maker_or_taker() == enums.ExchangeConstantsMarketPropertyColumns.MAKER.value



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.personal_data.transactions.types as transaction_types


def create_blockchain_transaction(exchange_manager, currency,
                                  blockchain_type, blockchain_transaction_id,
                                  blockchain_transaction_status=enums.BlockchainTransactionStatus.CREATED,
                                  source_address=None, destination_address=None,
                                  quantity=constants.ZERO, transaction_fee=constants.ZERO, is_deposit=False):
    blockchain_transaction = transaction_types.BlockchainTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.BLOCKCHAIN_DEPOSIT
        if is_deposit else enums.TransactionType.BLOCKCHAIN_WITHDRAWAL,
        currency=currency,
        blockchain_type=blockchain_type,
        blockchain_transaction_id=blockchain_transaction_id,
        blockchain_transaction_status=blockchain_transaction_status,
        source_address=source_address,
        destination_address=destination_address,
        quantity=quantity,
        transaction_fee=transaction_fee
    )
    _insert_transaction_instance(exchange_manager, blockchain_transaction)
    return blockchain_transaction


def create_realised_pnl_transaction(exchange_manager, currency, symbol, side,
                                    realised_pnl=constants.ZERO,
                                    is_closed_pnl=False,
                                    closed_quantity=constants.ZERO,
                                    cumulated_closed_quantity=constants.ZERO,
                                    first_entry_time=0,
                                    average_entry_price=constants.ZERO,
                                    average_exit_price=constants.ZERO,
                                    order_exit_price=constants.ZERO,
                                    leverage=constants.ZERO,
                                    trigger_source=enums.PNLTransactionSource.UNKNOWN):
    realised_pnl_transaction = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.CLOSE_REALISED_PNL
        if is_closed_pnl else enums.TransactionType.REALISED_PNL,
        currency=currency,
        symbol=symbol,
        realised_pnl=realised_pnl,
        closed_quantity=closed_quantity,
        cumulated_closed_quantity=cumulated_closed_quantity,
        first_entry_time=first_entry_time,
        average_entry_price=average_entry_price,
        average_exit_price=average_exit_price,
        order_exit_price=order_exit_price,
        leverage=leverage,
        trigger_source=trigger_source,
        side=side)
    _insert_transaction_instance(exchange_manager, realised_pnl_transaction)
    return realised_pnl_transaction


def create_fee_transaction(exchange_manager, currency, symbol,
                           quantity=constants.ZERO,
                           order_id=None,
                           funding_rate=constants.ZERO):
    fee_transaction = transaction_types.FeeTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.TRADING_FEE if order_id else enums.TransactionType.FUNDING_FEE,
        currency=currency,
        symbol=symbol,
        quantity=quantity,
        order_id=order_id,
        funding_rate=funding_rate)
    _insert_transaction_instance(exchange_manager, fee_transaction)
    return fee_transaction


def create_transfer_transaction(exchange_manager, currency, symbol):
    transfer_transaction = transaction_types.TransferTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        currency=currency,
        symbol=symbol)
    _insert_transaction_instance(exchange_manager, transfer_transaction)
    return transfer_transaction


def _insert_transaction_instance(exchange_manager, transaction):
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import pytest
import mock
import aiohttp
import ccxt.async_support as ccxt

import octobot_commons.aiohttp_util as aiohttp_util

import octobot_trading.exchanges as exchanges
import octobot_trading.exchanges.connectors.ccxt.ccxt_client_util as ccxt_client_util
import octobot_trading.constants as constants


@pytest.mark.asyncio
async def test_proxies():
    with mock.patch.object(aiohttp.ClientSession, "_request", mock.AsyncMock()) as _request_mock:

        # no proxy
        proxy_config = exchanges.ProxyConfig()
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] is None
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()

        # http proxy
        proxy_config = exchanges.ProxyConfig(
            http_proxy="http://127.0.0.1:9090",
        )
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] == "http://127.0.0.1:9090"
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()

        # http proxy callback
        proxy_config = exchanges.ProxyConfig(
            http_proxy_callback=lambda url, method, headers, body: "http://9.9.9.9:9090",
        )
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] == "http://9.9.9.9:9090"
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()

        # https proxy
        proxy_config = exchanges.ProxyConfig(
            https_proxy="https://127.0.0.1:9090",
        )
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] == "https://127.0.0.1:9090"
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()

        # https proxy callback
        proxy_config = exchanges.ProxyConfig(
            https_proxy_callback=lambda url, method, headers, body: "https://9.9.9.9:9090",
        )
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] == "https://9.9.9.9:9090"
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()

        # https proxy callback
        proxy_config = exchanges.ProxyConfig(
            https_proxy_callback=lambda url, method, headers, body: "https://9.9.9.9:9090",
        )
        async with _exchange_with_proxy_config(proxy_config) as exchange:
            _request_mock.assert_not_called()
            await exchange.load_markets()
            assert _request_mock.call_count > 0
            assert all(
                call.kwargs["proxy"] == "https://9.9.9.9:9090"
                for call in _request_mock.call_args_list
            )
            _request_mock.reset_mock()


@pytest.mark.asyncio
async def test_request_counter():
    with mock.patch.object(aiohttp.ClientSession, "_request", mock.AsyncMock()) as _request_mock:
        # without proxy
        proxy_config = exchanges.ProxyConfig()
        with mock.patch.object(constants, "ENABLE_CCXT_REQUESTS_COUNTER", True):
            async with _exchange_with_proxy_config(proxy_config, allow_request_counter=True) as exchange:
                await exchange.load_markets()
                assert _request_mock.call_count > 0
                assert all(
                    call.kwargs["proxy"] is None
                    for call in _request_mock.call_args_list
                )
                assert isinstance(exchange.session, aiohttp_util.CounterClientSession)
                # requests are counted by session counter
                assert len(exchange.session.per_min.paths) > 0
                assert exchange.session.per_min.paths[next(iter(exchange.session.per_min.paths))] > 0
        _request_mock.reset_mock()

        # with http proxy
        proxy_config = exchanges.ProxyConfig(http_proxy="http://127.0.0.1:9090")
        with mock.patch.object(constants, "ENABLE_CCXT_REQUESTS_COUNTER", True):
            async with _exchange_with_proxy_config(proxy_config, allow_request_counter=True) as exchange:
                await exchange.load_markets()
                assert _request_mock.call_count > 0
                assert all(
                    call.kwargs["proxy"] == "http://127.0.0.1:9090"
                    for call in _request_mock.call_args_list
                )
                assert isinstance(exchange.session, aiohttp_util.CounterClientSession)
                # requests are counted by session counter
                assert len(exchange.session.per_min.paths) > 0
                assert exchange.session.per_min.paths[next(iter(exchange.session.per_min.paths))] > 0
        _request_mock.reset_mock()

        # with http proxy callback
        proxy_config = exchanges.ProxyConfig(
            http_proxy_callback=lambda url, method, headers, body: "http://9.9.9.9:9090"
        )
        with mock.patch.object(constants, "ENABLE_CCXT_REQUESTS_COUNTER", True):
            async with _exchange_with_proxy_config(proxy_config, allow_request_counter=True) as exchange:
                await exchange.load_markets()
                assert _request_mock.call_count > 0
                assert all(
                    call.kwargs["proxy"] == "http://9.9.9.9:9090"
                    for call in _request_mock.call_args_list
                )
                assert isinstance(exchange.session, aiohttp_util.CounterClientSession)
                # requests are counted by session counter
                assert len(exchange.session.per_min.paths) > 0
                assert exchange.session.per_min.paths[next(iter(exchange.session.per_min.paths))] > 0
        _request_mock.reset_mock()


@contextlib.asynccontextmanager
async def _exchange_with_proxy_config(proxy_config: exchanges.ProxyConfig, allow_request_counter=False):
    exchange = None
    try:
        exchange = ccxt_client_util.instantiate_exchange(ccxt.kraken, {"enableRateLimit": False}, "test", proxy_config, allow_request_counter=allow_request_counter)
        yield exchange
    finally:
        if exchange:
            exchange.timeout_on_exit = 0    # avoid waiting for the exchange to close
            await exchange.close()



# pylint: disable=W0101
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import copy
import decimal
import time
import typing

import ccxt
import ccxt.pro as ccxtpro

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_trading.constants as trading_constants
import octobot_trading.enums as trading_enums
import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket_exchange
import octobot_trading.exchanges.connectors.ccxt.ccxt_client_util as ccxt_client_util
import octobot_trading.exchanges.connectors.ccxt.ccxt_adapter as ccxt_adapter
from octobot_trading.enums import (
    ExchangeConstantsOrderBookInfoColumns as ECOBIC,
    ExchangeConstantsTickersColumns as Ectc,
)
from octobot_trading.enums import WebsocketFeeds as Feeds


class CCXTWebsocketConnector(abstract_websocket_exchange.AbstractWebsocketExchange):
    INIT_REQUIRING_EXCHANGE_FEEDS = [Feeds.CANDLE]
    SUPPORTS_LIVE_PAIR_ADDITION = True
    FEED_INITIALIZATION_TIMEOUT = 15 * commons_constants.MINUTE_TO_SECONDS
    MIN_CONNECTION_CLOSE_INTERVAL = 2 * commons_constants.MINUTE_TO_SECONDS
    NO_MESSAGE_DISCONNECTED_TIMEOUT = 4 * commons_constants.MINUTE_TO_SECONDS
    RECREATE_CLIENT_ON_DISCONNECT = (
        False  # when True, a new ccxt websocket client will replace the previous
    )
    # one when the exchange is disconnected
    USE_REST_CONNECTOR_ADDITIONAL_CONFIG = False # if True, the additional config from the associated rest connector will be used
    FIX_CANDLES_TIMEZONE_IF_NEEDED: bool = False  # if True, the candles timestamps will be fixed to the UTC timezone, used when WS is returning candle timestamps in a non UTC timezone

    IGNORED_FEED_PAIRS = {
        # When ticker or future index is available : no need to calculate mark price from recent trades
        Feeds.TRADES: [Feeds.TICKER, Feeds.FUTURES_INDEX],
        # When candles are available : use min timeframe kline to push ticker
        Feeds.TICKER: [Feeds.KLINE],
        # When funding can be found in websocket ticker
        # Feeds.FUNDING: [Feeds.TICKER],
    }

    PAIR_INDEPENDENT_CHANNELS = [
        Feeds.PORTFOLIO,
        Feeds.ORDERS,
        Feeds.CREATE_ORDER,
        Feeds.CANCEL_ORDER,
        Feeds.TRADE,
        Feeds.LEDGER,
        Feeds.PORTFOLIO,
    ]
    WATCHED_PAIR_CHANNELS = [
        Feeds.TRADES,
        Feeds.TICKER,
        Feeds.CANDLE,
    ]
    CURRENT_TIME_FILTERED_CHANNELS = [
        Feeds.TRADES,
        Feeds.ORDERS,
        Feeds.TRADE,
    ]
    CANDLE_TIME_FILTERED_CHANNELS = [
        Feeds.CANDLE,
        Feeds.KLINE,
    ]
    # https://docs.ccxt.com/en/latest/ccxt.pro.manual.html?rtd_search=fetchLedger#real-time-vs-throttling
    # THROTTLED_CHANNELS are updated at each self.throttled_ws_updates.
    # Used as real time channels when self.throttled_ws_updates is 0
    # self.throttled_ws_updates is using trading_constants.THROTTLED_WS_UPDATES by default
    THROTTLED_CHANNELS = [
        Feeds.TICKER,
        Feeds.TRADES,
        Feeds.L1_BOOK,
        Feeds.L2_BOOK,
        Feeds.L3_BOOK,
    ]
    AUTHENTICATED_CHANNELS = [
        trading_enums.WebsocketFeeds.ORDERS,
        trading_enums.WebsocketFeeds.PORTFOLIO,
        trading_enums.WebsocketFeeds.TRADE,
        trading_enums.WebsocketFeeds.POSITION,
    ]
    EXCHANGE_CONSTRUCTOR_KWARGS = {}
    SHORT_RECONNECT_DELAY = 0.5
    LONG_RECONNECT_DELAY = 5

    def __init__(
        self,
        config,
        exchange_manager,
        adapter_class=None,
        additional_config=None,
        websocket_name=None,
    ):
        super().__init__(config, exchange_manager)
        self.filtered_pairs: list[str] = []
        self.watched_pairs: list[str] = []
        self.min_timeframe: typing.Optional[commons_enums.TimeFrames] = None
        self._previous_open_candles: dict[str, dict[str, dict]] = {}
        self._subsequent_unordered_candles_count: dict[
            str, dict[str, tuple[int, float]]
        ] = {}  # dict values: tuple(candle_count, candle_time)
        self._errors_count: dict[str, dict[str, int]] = {}
        self._start_time_millis: typing.Optional[int] = (
            None  # used for the "since" param in CURRENT/CANDLE_TIME_FILTERED_CHANNELS
        )
        self.websocket_name: str = websocket_name or self.get_name()

        self.local_loop: typing.Optional[asyncio.AbstractEventLoop] = None

        self.should_stop: bool = False
        self.is_authenticated: bool = False
        self.adapter: ccxt_adapter.CCXTAdapter = self.get_adapter_class(adapter_class)(
            self
        )
        self.additional_config: typing.Optional[dict[str, typing.Any]] = (
            additional_config
        )
        if self.USE_REST_CONNECTOR_ADDITIONAL_CONFIG:
            self.additional_config = {
                **(self.additional_config or {}),
                **self.exchange_manager.exchange.get_extended_additional_connector_config()
            }
        self.headers: dict[str, str] = {}
        self.options: dict[str, typing.Any] = {
            "newUpdates": True,  # only get new updates from trades and ohlcv (don't return the full cached history)
            "OHLCVLimit": trading_constants.CCXT_OHLCV_CACHE_LIMIT,
            "tradesLimit": trading_constants.CCXT_TRADES_CACHE_LIMIT,
            "ordersLimit": trading_constants.CCXT_ORDERS_CACHE_LIMIT,
            "watchOrderBookLimit": trading_constants.CCXT_WATCH_ORDER_BOOK_LIMIT,
        }
        # add default options
        self.add_options(
            ccxt_client_util.get_ccxt_client_login_options(self.exchange_manager)
        )
        self.client: ccxt.Exchange = None  # type: ignore # ccxt.pro exchange: a ccxt.async_support exchange with websocket capabilities
        self.feed_tasks: dict[str, asyncio.Task] = {}
        self._last_close_time: float = 0
        self._last_message_time: float = 0
        self.throttled_ws_updates: float = trading_constants.THROTTLED_WS_UPDATES

        self.candle_delay_seconds: float = 0

        self._create_client()

    """
    Methods
    """

    def get_feed_name(self):
        return self.websocket_name

    def start(self):
        asyncio.run(self._inner_start())

    async def _inner_start(self):
        self._start_time_millis = self.client.milliseconds()
        self.stopped_event = asyncio.Event()
        self.local_loop = asyncio.get_event_loop()
        await self._init_client()
        # keep loop open till stop is called
        await self.stopped_event.wait()

    def get_adapter_class(self, adapter_class):
        return adapter_class or ccxt_adapter.CCXTAdapter

    def add_headers(self, headers_dict):
        """
        Add new headers to ccxt client
        :param headers_dict: the additional header keys and values as dict
        """
        self.headers.update(headers_dict)
        if self.client is not None:
            ccxt_client_util.add_headers(self.client, headers_dict)

    def add_options(self, options_dict):
        """
        Add new options to ccxt client
        :param options_dict: the additional option keys and values as dict
        """
        self.options.update(options_dict)
        if self.client is not None:
            ccxt_client_util.add_options(self.client, options_dict)

    async def _inner_stop(self):
        self.should_stop = True
        try:
            if self.client is not None:
                await self.client.close()
            self.stopped_event.set()
        except Exception as e:
            self.logger.exception(e, False)
            self.logger.error(f"Failed to stop websocket feed : {e}")

    async def stop(self):
        """
        Reimplementation of self.client.stop() without calling loop.run_until_complete()
        """
        if asyncio.get_event_loop() is self.local_loop:
            await self._inner_stop()
        else:
            asyncio_tools.run_coroutine_in_asyncio_loop(
                self._inner_stop(), self.local_loop
            )

    async def close(self):
        """
        Can't call self.client.close() because it uses loop operations
        """

    def add_pairs(self, pairs, watching_only=False):
        """
        Add new pairs to self.filtered_pairs
        :param pairs: the list of pair to add
        :param watching_only: if pairs are for watching or trading purpose
        """
        for pair in pairs:
            self._add_pair(pair, watching_only=watching_only)

    def remove_pairs(self, pairs, watching_only=False):
        """
        Remove pairs from self.filtered_pairs
        :param pairs: the list of pair to remove
        :param watching_only: if pairs are for watching or trading purpose
        """
        for pair in pairs:
            self._remove_pair(pair, watching_only=watching_only)

    async def _inner_update_followed_pairs(self):
        if self.initialized_event is None:
            # should never happen
            return
        if not self.initialized_event.is_set():
            await self.initialized_event.wait()
        self._filter_exchange_symbols()
        self._subscribe_channels_feeds(True)

    def update_followed_pairs(self):
        asyncio_tools.run_coroutine_in_asyncio_loop(
            self._inner_update_followed_pairs(), self.local_loop
        )

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:
        return cls.get_name() == exchange_candidate_name

    def is_time_frame_supported(self, time_frame):
        return time_frame.value in ccxt_client_util.get_time_frames(self.client)

    def get_pair_from_exchange(self, pair):
        """
        Convert a ccxt symbol format to uniformized symbol format
        :param pair: the pair to format
        :return: the formatted pair when success else an empty string
        """
        # octobot uses the ccxt format for pairs
        return pair

    def get_exchange_pair(self, pair) -> str:
        """
        Convert an uniformized symbol format to a ccxt symbol format
        :param pair: the pair to format
        :return: the ccxt pair when success else an empty string
        """
        # octobot uses the ccxt format for pairs
        return pair

    """
    Private methods
    """

    def _create_client(self):
        """
        Creates ccxt client instance
        """
        client_class = getattr(ccxtpro, self.get_feed_name())
        self.client, self.is_authenticated = ccxt_client_util.create_client(
            client_class,
            self.exchange_manager,
            self.logger,
            self.options,
            self.headers,
            self.additional_config,
            self._should_authenticate(),
            keys_adapter=self._get_keys_adapter(),
            # Always disable request counter: ws exchange will be started in a different async loop,
            # which is not yet supported
            allow_request_counter=False,
        )
        if self.exchange_manager.exchange.is_supporting_sandbox():
            ccxt_client_util.set_sandbox_mode(self, self.exchange_manager.is_sandboxed)
    
    def _get_keys_adapter(self):
        return None

    def _should_authenticate(self):
        if self.exchange_manager.exchange.requires_authentication(
            self.exchange_manager.exchange.tentacle_config, None, None
        ):
            return True
        return self._has_authenticated_channel() and super()._should_authenticate()

    def _has_authenticated_channel(self) -> bool:
        for feed in self.AUTHENTICATED_CHANNELS:
            if feed in self.EXCHANGE_FEEDS:
                return True
        return False

    def _is_authenticated_feed(self, feed):
        return feed in self.AUTHENTICATED_CHANNELS

    async def _init_client(self):
        """
        Prepares client configuration and instantiates client feeds
        """
        try:
            self.initialized_event = asyncio.Event()
            self.logger.info(
                f"Loading {self.exchange_manager.exchange_name} websocket exchange markets"
            )
            await self.client.load_markets()
            self._filter_exchange_pairs_and_timeframes()
            self._subscribe_feeds()
        except Exception as e:
            self.logger.exception(
                e, True, f"Failed to subscribe when creating websocket feed : {e}"
            )
        finally:
            self.initialized_event.set()

    async def _close_exchange_to_force_reconnect(self):
        if (
            time.time() - self._last_close_time > self.MIN_CONNECTION_CLOSE_INTERVAL
            and not self.should_stop
        ):
            # Close client to force connections re-open. The next watch_xyz will recreate the connection
            self._last_close_time = time.time()
            self.logger.debug(f"Closing exchange connection.")
            await self.client.close()
            if self.RECREATE_CLIENT_ON_DISCONNECT:
                self._create_client()
            self.logger.debug(
                "Exchange connection closed. The next watch-xyz will re-open it."
            )
            return True
        return False

    def _subscribe_feeds(self):
        """
        Prepares the subscription of unauthenticated and authenticated feeds and subscribe all
        """
        if self._should_run_candle_feed():
            if self.min_timeframe is None:
                self.logger.error(
                    f"Missing min_timeframe in exchange websocket with candle feeds. This probably means that no "
                    f"required time frame is supported by this exchange's websocket "
                    f"(valid_candle_intervals: {ccxt_client_util.get_time_frames(self.client)})"
                )
            self._subscribe_candle_feed()

        # drop unsupported channels
        self.channels = [
            channel
            for channel in self.channels
            if self._is_supported_channel(channel)
            and channel != self.EXCHANGE_FEEDS.get(Feeds.CANDLE)
        ]

        self._subscribe_channels_feeds(False)

    def _should_use_authenticated_feeds(self):
        """
        :return: True when authenticated feeds shouldn't be added
        """
        return (
            self._should_authenticate()
            and self.exchange.authenticated()
            and not self.exchange_manager.is_sandboxed
        )

    def _is_supported_channel(self, channel):
        """
        Checks if the channel is supported
        :param channel: the channel name
        :return: True if the channel is not unsupported, not ignored
        and if it's an authenticated channel if the exchange is authenticated
        """
        if (
            self._is_authenticated_feed(channel)
            and not self._should_use_authenticated_feeds()
        ):
            return False
        return not (
            self.EXCHANGE_FEEDS.get(channel, Feeds.UNSUPPORTED.value)
            == Feeds.UNSUPPORTED.value
            or self.should_ignore_feed(channel)
        )

    @classmethod
    def get_exchange_feed(cls, feed) -> str:
        feed_value = cls.EXCHANGE_FEEDS.get(
            feed, trading_enums.WebsocketFeeds.UNSUPPORTED.value
        )
        if cls.is_feed_supported(feed_value):
            return feed.value
        return trading_enums.WebsocketFeeds.UNSUPPORTED.value

    def _subscribe_candle_feed(self):
        """
        Subscribes a new candle feed for each time frame
        """
        for time_frame in self.time_frames:
            self._subscribe_feed(
                Feeds.CANDLE,
                symbols=self.filtered_pairs,
                time_frame=time_frame.value,
            )

    def _subscribe_channels_feeds(self, pairs_related_channels_only):
        """
        Subscribes all time frame unrelated feeds
        """
        if not pairs_related_channels_only:
            self._subscribe_pair_independent_feed()
        self._subscribe_traded_pairs_feed()
        self._subscribe_watched_pairs_feed()

    def _subscribe_pair_independent_feed(self):
        """
        Subscribes all pair unrelated feeds
        """
        feeds = [
            channel
            for channel in self.channels
            if self._is_pair_independent_feed(channel)
        ]
        for feed in feeds:
            self._subscribe_feed(
                feed,
            )

    def _subscribe_traded_pairs_feed(self):
        """
        Subscribes all time frame unrelated feeds for traded pairs
        """
        if not self.filtered_pairs:
            return
        feeds = [
            channel
            for channel in self.channels
            if not self._is_pair_independent_feed(channel)
        ]
        for feed in feeds:
            self._subscribe_feed(
                feed,
                symbols=self.filtered_pairs,
            )

    def _subscribe_watched_pairs_feed(self):
        """
        Subscribes feeds for watched pairs (only on one timeframe for multiple timeframes feeds)
        """
        if not self.watched_pairs:
            return
        feeds = [
            channel
            for channel in self.WATCHED_PAIR_CHANNELS
            if self._is_supported_channel(channel)
            and (channel in self.channels or not self.channels)
        ]
        for feed in feeds:
            self._subscribe_feed(
                feed,
                symbols=self.watched_pairs,
            )

    def _get_feed_default_kwargs(self):
        return {}

    def _get_feed_generator_by_feed(self):
        return {
            # Subscribed feeds
            # Unauthenticated
            Feeds.TRADES: self._get_generator("watchTrades"),
            Feeds.TICKER: self._get_generator("watchTicker"),
            Feeds.CANDLE: self._get_generator("watchOHLCV"),
            Feeds.KLINE: self._get_generator("watchOHLCV"),
            Feeds.L1_BOOK: self._get_generator("watchOrderBook"),
            Feeds.L2_BOOK: self._get_generator("watchOrderBook"),
            Feeds.L3_BOOK: self._get_generator("watchOrderBook"),
            Feeds.MARKETS: self._get_generator("watchMarkets"),
            Feeds.FUNDING: Feeds.UNSUPPORTED,
            Feeds.LIQUIDATIONS: Feeds.UNSUPPORTED,
            Feeds.OPEN_INTEREST: Feeds.UNSUPPORTED,
            Feeds.FUTURES_INDEX: Feeds.UNSUPPORTED,
            # Authenticated
            Feeds.TRANSACTIONS: self._get_generator("watchTransactions"),
            Feeds.PORTFOLIO: self._get_generator("watchBalance"),
            Feeds.ORDERS: self._get_generator("watchOrders"),
            Feeds.TRADE: self._get_generator("watchMyTrades"),
            Feeds.LEDGER: self._get_generator("watchLedger"),
            Feeds.POSITION: Feeds.UNSUPPORTED,
            # Publish feeds
            Feeds.CREATE_ORDER: self._get_generator("watchCreateOrder"),
            Feeds.CANCEL_ORDER: self._get_generator("watchCancelOrder"),
        }

    def _get_generator(self, method_name):
        return (
            getattr(self.client, method_name)
            if hasattr(self.client, method_name)
            and self.client.has.get(method_name, False)
            else Feeds.UNSUPPORTED
        )

    def _get_callback_by_feed(self):
        return {
            # Unauthenticated
            Feeds.TRADES: self.recent_trades,
            Feeds.TICKER: self.ticker,
            Feeds.CANDLE: self.candle,
            Feeds.KLINE: self.candle,
            Feeds.FUNDING: self.funding,
            Feeds.OPEN_INTEREST: self.open_interest,
            Feeds.L1_BOOK: self.book,
            Feeds.L2_BOOK: self.book,
            Feeds.L3_BOOK: self.book,
            Feeds.MARKETS: self.markets,
            # Authenticated
            Feeds.TRANSACTIONS: self.transaction,
            Feeds.PORTFOLIO: self.balance,
            Feeds.ORDERS: self.orders,
            Feeds.TRADE: self.trades,
        }

    def _get_since_filter_value(self, feed, time_frame):
        if feed in self.CURRENT_TIME_FILTERED_CHANNELS:
            return self._start_time_millis
        elif feed in self.CANDLE_TIME_FILTERED_CHANNELS:
            candles_ms = (
                commons_enums.TimeFramesMinutes[commons_enums.TimeFrames(time_frame)]
                * commons_constants.MSECONDS_TO_MINUTE
            )
            time_delta = self._start_time_millis % candles_ms
            return self._start_time_millis - time_delta
        return None

    def _subscribe_feed(
        self, feed, symbols=None, time_frame=None, since=None, limit=None, params=None
    ):
        """
        Subscribe a new feed
        :param feed: the feed to subscribe to
        :param symbols: the feed symbols
        :param time_frame: the feed time_frame
        :param since: the ccxt feed since arg
        :param limit: the ccxt feed limit arg
        :param params: the ccxt feed param arg
        """
        try:
            feed_callback = self._get_callback_by_feed()[feed]
            feed_generator = self._get_feed_generator_by_feed()[feed]
            if Feeds.UNSUPPORTED in (feed_callback, feed_generator):
                raise KeyError
        except KeyError:
            self.logger.error(f"Impossible to subscribe to {feed}: feed not supported")
            return
        added_subscriptions = []
        has_added_feed = False
        if feed in self.TIME_FRAME_RELATED_FEEDS and time_frame is None:
            time_frame = self.min_timeframe.value
        kwargs = copy.copy(self._get_feed_default_kwargs())
        if time_frame is not None:
            kwargs["timeframe"] = time_frame
        if since is not None:
            kwargs["since"] = since
        else:
            auto_since = self._get_since_filter_value(feed, time_frame)
            if auto_since is not None:
                kwargs["since"] = auto_since
            elif feed in self.CURRENT_TIME_FILTERED_CHANNELS:
                kwargs["since"] = self._start_time_millis
            elif feed in self.CANDLE_TIME_FILTERED_CHANNELS:
                kwargs["since"] = self._start_time_millis
        if limit is not None:
            kwargs["limit"] = limit
        if params is not None:
            kwargs["params"] = params
        if symbols is not None:
            for symbol in symbols:
                kwargs["symbol"] = symbol
                # one task per symbol: ccxt_pro is not handling multi symbol generators
                if self._create_task_if_necessary(
                    feed, feed_callback, feed_generator, **kwargs
                ):
                    added_subscriptions.append(symbol)
                    has_added_feed = True
        else:
            # no symbol param
            if self._create_task_if_necessary(
                feed, feed_callback, feed_generator, **kwargs
            ):
                has_added_feed = True

        symbols_str = (
            f"for {', '.join(added_subscriptions)} " if added_subscriptions else ""
        )
        time_frame_str = f"on {time_frame}" if time_frame else ""
        if has_added_feed:
            self.logger.debug(
                f"Subscribed to {feed.value} {symbols_str}{time_frame_str}"
            )
        else:
            self.logger.debug(
                f"No new feed to subscribe to on {feed.value} (inputs: {symbols_str}{time_frame_str})"
            )

    async def _feed_task(self, feed, callback, watch_func, *g_args, **g_kwargs):
        if not await self._wait_for_initialization(feed, *g_args, **g_kwargs):
            self.logger.error(
                f"Aborting {feed.value} feed connection with {g_kwargs}: "
                f"missing required initialization data"
            )
            return
        enable_throttling = (
            feed in self.THROTTLED_CHANNELS and self.throttled_ws_updates != 0.0
        )
        ws_des = f"{watch_func.__name__} {g_kwargs}"
        subsequent_disconnections = 0
        already_got_feed_stopping_error = False
        already_got_closed_by_user_error = False
        spamming_logs_warning_interval = 5000
        spamming_logs_debug_interval = 1000
        while not self.should_stop:
            try:
                update_data = await watch_func(*g_args, **g_kwargs)

                self._last_message_time = time.time()
                if subsequent_disconnections > 0:
                    self.logger.debug(f"Reconnected to {ws_des}")
                subsequent_disconnections = 0
                already_got_closed_by_user_error = False
                if update_data:
                    # Use a copy of the update data as it will be edited by adapters.
                    # We should avoid editing the original object since it is also used in ccxt internally buffers
                    await callback(copy.deepcopy(update_data), **g_kwargs)
                if enable_throttling:
                    # ccxt keeps updating the internal structures while waiting
                    # https://docs.ccxt.com/en/latest/ccxt.pro.manual.html?rtd_search=fetchLedger#incremental-data-structures
                    await asyncio.sleep(self.throttled_ws_updates)
            except ccxt.NetworkError as err:
                # short reconnect on ping pong timeout or 1st reconnect
                is_ping_pong_error = isinstance(err, ccxt.RequestTimeout)
                if (
                    is_ping_pong_error
                    and time.time() - self._last_message_time
                    < self.NO_MESSAGE_DISCONNECTED_TIMEOUT
                ):
                    # last message not received long ago: the websocket is probably not disconnected, it's just a
                    # pong timeout, which can happen: there is no issue, instantly reconnect feed
                    self.logger.debug(
                        f"Instantly reconnecting {ws_des} feed after ping-pong timeout."
                    )
                    continue
                reconnect_delay = (
                    self.SHORT_RECONNECT_DELAY
                    if (subsequent_disconnections == 0 or is_ping_pong_error)
                    else self.LONG_RECONNECT_DELAY
                )
                self.logger.debug(
                    f"Can't connect to exchange {ws_des} websocket: {err}. "
                    f"Retrying in {reconnect_delay} seconds"
                )
                if await self._close_exchange_to_force_reconnect():
                    message = (
                        f"Closed exchange connection to force reconnect. Error: {err}"
                    )
                    if (
                        subsequent_disconnections > 1
                        and subsequent_disconnections % 5 == 0
                    ):
                        self.logger.error(
                            f"Multiple disconnections in a row [{subsequent_disconnections}] for {ws_des}. {message}"
                        )
                    else:
                        self.logger.debug(message)
                await asyncio.sleep(reconnect_delay)
                self.logger.debug(f"Reconnecting to {ws_des}")
                # self.client might have changed
                watch_func = self._get_feed_generator_by_feed()[feed]
                subsequent_disconnections += (
                    1  # wait for a longer time before the next reconnect
                )
            except ccxt.BadRequest as err:
                message = f"Impossible to start {ws_des} feed due to exchange refusing the connection request: {err}."
                self.logger.exception(
                    err,
                    True,
                    f"{message} {'Will retry once' if not already_got_feed_stopping_error else 'Now stopping'}.",
                )
                if already_got_feed_stopping_error:
                    # there is a real issue when connecting to the feed. Don't loop
                    return
                already_got_feed_stopping_error = True
                await asyncio.sleep(self.LONG_RECONNECT_DELAY)  # avoid spamming
                # self.client might have changed
                watch_func = self._get_feed_generator_by_feed()[feed]
            except ccxt.NotSupported as err:
                self.logger.exception(
                    err,
                    True,
                    f"Impossible to start {ws_des} feed: {err}. "
                    f"Stopping it. Please report to the OctoBot team if you see this error",
                )
                return
            except Exception as err:
                count_error = True
                if isinstance(err, ccxt.ExchangeClosedByUser):
                    if self.should_stop:
                        # normal when stopping websocket
                        self.logger.debug(
                            f"Disconnected {ws_des}: connection closed as {self.get_name()} is stopping ({err})"
                        )
                        return
                    # happening when auto-stopping websocket (calling self._close_exchange_to_force_reconnect)
                    # from another task: current task also raised ccxt.ExchangeClosedByUser
                    # this is normal as long as it happens only once after a reconnection
                    if not already_got_closed_by_user_error:
                        count_error = False
                        self.logger.debug(
                            f"{ws_des} connection automatically closed, reconnecting in {self.LONG_RECONNECT_DELAY}"
                            f" seconds ({err})"
                        )
                        already_got_closed_by_user_error = True
                if count_error:
                    error_count = self._increment_error_counter(
                        g_kwargs.get("time_frame"), err
                    )
                    error_message = (
                        f"Unexpected error when handling {ws_des} feed: {err} "
                        f"({err.__class__.__name__}) ({error_count} times)"
                    )
                    if error_count == 1:
                        self.logger.exception(err, True, error_message)
                    elif error_count % spamming_logs_warning_interval == 0:
                        self.logger.warning(error_message)
                    elif error_count % spamming_logs_debug_interval == 0:
                        self.logger.debug(error_message)
                await asyncio.sleep(self.LONG_RECONNECT_DELAY)  # avoid spamming
                subsequent_disconnections += (
                    1  # wait for a longer time before the next reconnect
                )
                # self.client might have changed
                watch_func = self._get_feed_generator_by_feed()[feed]

    def _create_task_if_necessary(self, feed, feed_callback, feed_generator, **kwargs):
        identifier = self._get_feed_identifier(feed_generator, kwargs)
        if identifier not in self.feed_tasks:
            self.logger.debug(
                f"Subscribing to {feed.value} with {kwargs} ({len(self.feed_tasks)} total feeds)"
            )
            self.feed_tasks[identifier] = asyncio.create_task(
                self._feed_task(feed, feed_callback, feed_generator, **kwargs)
            )
            return True
        return False

    async def _wait_for_initialization(self, feed, *g_args, **g_kwargs):
        if (
            not self.is_feed_requiring_init(feed)
            or g_kwargs["symbol"] not in self.filtered_pairs
        ):
            # no need to wait for pairs not in self.filtered_pairs
            return True
        is_initialized_func = None
        if feed is Feeds.CANDLE:

            def candle_is_initialized_func():
                if self.exchange_manager is None:
                    # Should only happen in tests / unusual environments. Or there is a real issue.
                    self.logger.error(
                        f"No exchange manager when starting websocket connector."
                    )
                    return False
                try:
                    return (
                        self.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
                            g_kwargs["symbol"], allow_creation=False
                        )
                        .symbol_candles[commons_enums.TimeFrames(g_kwargs["timeframe"])]
                        .candles_initialized
                    )
                except KeyError:
                    return False

            is_initialized_func = candle_is_initialized_func
        if is_initialized_func is None:
            return True
        if is_initialized_func():
            return True
        self.logger.debug(
            f"Waiting for initialization before starting {feed.value} feed with {g_kwargs}"
        )
        t0 = time.time()
        while (
            not self.should_stop and time.time() - t0 < self.FEED_INITIALIZATION_TIMEOUT
        ):
            # add timeout
            if is_initialized_func():
                self.logger.debug(
                    f"Starting {feed.value} feed with {g_kwargs}: initialization complete"
                )
                return True
            # quickly update at first
            await asyncio.sleep(
                0.1 if time.time() - t0 < self.FEED_INITIALIZATION_TIMEOUT / 10 else 1
            )
        return is_initialized_func()

    def _get_feed_identifier(self, feed_generator, kwargs):
        return f"{feed_generator.__name__}{kwargs}"

    def _filter_exchange_pairs_and_timeframes(self):
        """
        Populates self.filtered_pairs and self.min_timeframe
        """
        self._add_exchange_symbols()
        self._init_exchange_time_frames()

    def _add_pair(self, pair, watching_only):
        """
        Add a pair to self.filtered_pairs if supported
        :param pair: the pair to add
        :param watching_only: when True add pair to watched_pairs else to filtered_pairs
        """
        if watching_only:
            if pair not in self.watched_pairs:
                self.watched_pairs.append(pair)
        else:
            if pair not in self.filtered_pairs:
                self.filtered_pairs.append(pair)

    def _remove_pair(self, pair, watching_only):
        """
        Remove a pair from self.filtered_pairs if supported
        :param pair: the pair to remove
        :param watching_only: when True remove pair from watched_pairs else from filtered_pairs
        """
        if watching_only:
            if pair in self.watched_pairs:
                self.watched_pairs.remove(pair)
        else:
            if pair in self.filtered_pairs:
                self.filtered_pairs.remove(pair)

    def _add_exchange_symbols(self):
        """
        Populates self.filtered_pairs from self.pairs when pair is supported by the ccxt exchange
        """
        for pair in self.pairs:
            self._add_pair(pair, watching_only=False)
        self._filter_exchange_symbols()

    def _filter_exchange_symbols(self):
        pre_filter_watched_pairs = copy.copy(self.watched_pairs)
        self.watched_pairs = [
            pair for pair in pre_filter_watched_pairs if self._is_supported_pair(pair)
        ]
        pre_filter_filtered_pairs = copy.copy(self.filtered_pairs)
        self.filtered_pairs = [
            pair for pair in pre_filter_filtered_pairs if self._is_supported_pair(pair)
        ]
        unsupported_pairs = []
        for pairs, pre_filter_pairs in (
            (self.watched_pairs, pre_filter_watched_pairs),
            (self.filtered_pairs, pre_filter_filtered_pairs),
        ):
            if len(pre_filter_filtered_pairs) > len(pairs):
                unsupported_pairs += [
                    pair for pair in pre_filter_pairs if pair not in pairs
                ]
        if unsupported_pairs:
            self.logger.error(
                f"{unsupported_pairs} pair is not supported by this exchange's websocket"
            )

    def _add_time_frame(self, filtered_timeframes, time_frame, log_on_error):
        """
        Add a time frame to filtered_timeframes if supported
        :param time_frame: the time frame to add
        """
        if self.is_time_frame_supported(time_frame):
            filtered_timeframes.append(time_frame)
        elif log_on_error:
            self.logger.error(
                f"{time_frame.value} time frame is not supported by this exchange's websocket"
            )

    def _init_exchange_time_frames(self):
        """
        Populates self.min_timeframe from self.time_frames when time frame is supported by the ccxt exchange
        Leaves self.min_timeframe at None if no timeframe is available on this exchange ws
        """
        # Log error only if necessary (self.min_timeframe is used by candle feed only)
        log_on_error = self._should_run_candle_feed()
        filtered_timeframes = []
        for time_frame in self.time_frames:
            self._add_time_frame(filtered_timeframes, time_frame, log_on_error)
        if filtered_timeframes:
            self.min_timeframe = time_frame_manager.find_min_time_frame(
                filtered_timeframes
            )

    def _should_run_candle_feed(self):
        return (
            self.EXCHANGE_FEEDS.get(Feeds.CANDLE, Feeds.UNSUPPORTED.value)
            != Feeds.UNSUPPORTED.value
        )

    def _is_supported_pair(self, pair):
        return pair in ccxt_client_util.get_symbols(self.client, True)

    def _is_pair_independent_feed(self, feed):
        return feed in self.PAIR_INDEPENDENT_CHANNELS

    @classmethod
    def get_feeds_count(cls, pairs, time_frames) -> int:
        total_count = 0
        if pairs:
            # rounding: account for 1 feed per symbol
            total_count = len(pairs)
            if cls.get_exchange_feed(trading_enums.WebsocketFeeds.CANDLE):
                # rounding: account for 1 per timeframe per symbol
                total_count += len(pairs) * len(time_frames)
        return total_count

    def _convert_book_prices_to_orders(self, book_prices_and_volumes, book_side):
        """
        Convert a book_prices format : {PRICE_1: SIZE_1, PRICE_2: SIZE_2...}
        to OctoBot's order book format
        :param book_prices: an order book dictionary (order_book.SortedDict)
        :param book_side: a TradeOrderSide value
        :return: the list of order book data converted
        """
        return [
            {
                ECOBIC.PRICE.value: float(book_price_and_volume[0]),
                ECOBIC.SIZE.value: float(book_price_and_volume[1]),
                ECOBIC.SIDE.value: book_side,
            }
            for book_price_and_volume in book_prices_and_volumes
        ]

    """
    Callbacks
    """

    async def ticker(self, ticker: dict, symbol=None, **kwargs):
        """
        :param ticker: the ccxt ticker dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        adapted = self.adapter.adapt_ticker(ticker)
        await self.push_to_channel(
            trading_constants.TICKER_CHANNEL,
            symbol,
            adapted,
        )

    async def recent_trades(self, trades: list, symbol=None, **kwargs):
        """
        :param trades: the ccxt ticker list
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        adapted = self.adapter.adapt_public_recent_trades(trades)
        await self.push_to_channel(
            trading_constants.RECENT_TRADES_CHANNEL, symbol, adapted
        )

    async def book(self, order_book: dict, symbol=None, **kwargs):
        """
        :param order_book: the ccxt order_book dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        book_instance = self.get_book_instance(symbol)

        book_instance.handle_book_adds(
            self._convert_book_prices_to_orders(
                order_book[ECOBIC.ASKS.value], trading_enums.TradeOrderSide.SELL.value
            )
            + self._convert_book_prices_to_orders(
                order_book[ECOBIC.BIDS.value], trading_enums.TradeOrderSide.BUY.value
            )
        )

        await self.push_to_channel(
            trading_constants.ORDER_BOOK_CHANNEL,
            symbol,
            book_instance.asks,
            book_instance.bids,
            update_order_book=False,
        )

    async def markets(self, markets: list, symbol=None, **kwargs):
        """
        :param market: the ccxt market dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        raise NotImplementedError("markets callback is not implemented")

    async def candle(self, candles: list, symbol=None, timeframe=None, **kwargs):
        """
        :param candles: the ccxt ohlcv list
        :param symbol: the feed symbol
        :param timeframe: the feed timeframe
        :param kwargs: the feed kwargs
        """
        time_frame = commons_enums.TimeFrames(timeframe)
        if self.FIX_CANDLES_TIMEZONE_IF_NEEDED:
            try:
                candles = self._fix_candles_timestamps(candles, time_frame)
            except ValueError as err:
                self.logger.info(f"Skipped websocket candle({symbol} {timeframe}): {err}")
                return
        kline = self.adapter.adapt_kline([copy.deepcopy(candles[-1])])[0]
        adapted = self.adapter.adapt_ohlcv(candles, time_frame=time_frame)
        last_candle = adapted[-1]
        if symbol not in self.watched_pairs:
            for candle in adapted:
                previous_candle = self._get_previous_open_candle(timeframe, symbol)
                is_previous_candle_closed = False
                if previous_candle is not None:
                    current_candle_time = candle[
                        commons_enums.PriceIndexes.IND_PRICE_TIME.value
                    ]
                    previous_candle_time = previous_candle[
                        commons_enums.PriceIndexes.IND_PRICE_TIME.value
                    ]
                    if previous_candle_time < current_candle_time:
                        # new candle is after the previous one: the previous one is now closed
                        is_previous_candle_closed = True
                    elif previous_candle_time > current_candle_time:
                        # should not happen: exchange feed is providing past candles after newer ones
                        # candle feed should be marked as unsupported in this exchange (at least for now)

                        # update internal candle store to still keep track of the most up to date candle
                        await self.exchange_manager.get_symbol_data(
                            symbol
                        ).handle_candles_update(
                            time_frame,
                            candle,
                            replace_all=False,
                            partial=False,
                            upsert=True,
                        )
                        is_new_unordered_candle = (
                            self._register_subsequent_unordered_candle(
                                timeframe, symbol, time_frame, current_candle_time
                            )
                        )
                        subsequent_unordered_candles = (
                            self._get_subsequent_unordered_candles_count(
                                timeframe, symbol
                            )
                        )
                        error_message = (
                            f"Ignored unexpected candle for {symbol} on {timeframe}: "
                            f"candle time {current_candle_time}, "
                            f"previous candle time: {previous_candle_time} "
                            f"({subsequent_unordered_candles} unordered candles in a row)."
                        )
                        if is_new_unordered_candle:
                            self.logger.warning(error_message)
                        # prevent spamming
                        elif subsequent_unordered_candles < 3 or (
                            subsequent_unordered_candles % 1000 == 0
                        ):
                            self.logger.debug(error_message)
                        if candle is last_candle:
                            # last candle in loop: don't go any further
                            return
                        else:
                            # go to next candle in loop
                            continue
                if is_previous_candle_closed:
                    # OHLCV_CHANNEL only takes closed candles
                    await self.push_to_channel(
                        trading_constants.OHLCV_CHANNEL,
                        time_frame,
                        symbol,
                        previous_candle,
                    )
                self._register_previous_open_candle(timeframe, symbol, candle)
            await self.push_to_channel(
                trading_constants.KLINE_CHANNEL, time_frame, symbol, kline
            )

        # Push a new ticker if necessary : only push on the min timeframe
        if time_frame is self.min_timeframe:
            ticker = {
                Ectc.HIGH.value: kline[commons_enums.PriceIndexes.IND_PRICE_HIGH.value],
                Ectc.LOW.value: kline[commons_enums.PriceIndexes.IND_PRICE_LOW.value],
                Ectc.BID.value: None,
                Ectc.BID_VOLUME.value: None,
                Ectc.ASK.value: None,
                Ectc.ASK_VOLUME.value: None,
                Ectc.OPEN.value: kline[commons_enums.PriceIndexes.IND_PRICE_OPEN.value],
                Ectc.CLOSE.value: kline[
                    commons_enums.PriceIndexes.IND_PRICE_CLOSE.value
                ],
                Ectc.LAST.value: kline[
                    commons_enums.PriceIndexes.IND_PRICE_CLOSE.value
                ],
                Ectc.PREVIOUS_CLOSE.value: None,
                # do not include volume as this is not the ticker volume but the current candle volume
                Ectc.BASE_VOLUME.value: None,
                Ectc.TIMESTAMP.value: self.exchange.get_exchange_current_time(),
            }
            await self.push_to_channel(trading_constants.TICKER_CHANNEL, symbol, ticker)

    async def funding(self, funding: dict, symbol=None, **kwargs):
        """
        Unsupported, feed list https://docs.ccxt.com/en/latest/ccxt.pro.manual.html?rtd_search=fetchLedger#prerequisites
        :param funding: the ccxt funding dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported
        raise NotImplementedError("funding callback is not implemented")
        adapted = self.adapter.parse_funding_rate(funding)
        predicted_funding_rate = adapted.get(
            trading_enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value,
            trading_constants.NaN,
        )
        await self.push_to_channel(
            trading_constants.FUNDING_CHANNEL,
            symbol,
            decimal.Decimal(
                adapted[
                    trading_enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value
                ]
            ),
            predicted_funding_rate=decimal.Decimal(
                str(predicted_funding_rate or trading_constants.NaN)
            ),
            next_funding_time=adapted[
                trading_enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value
            ],
            last_funding_time=adapted[
                trading_enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value
            ],
        )

    async def open_interest(self, open_interest: dict, symbol=None, **kwargs):
        """
        Unsupported, feed list https://docs.ccxt.com/en/latest/ccxt.pro.manual.html?rtd_search=fetchLedger#prerequisites
        :param open_interest: the ccxt open_interest dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported
        raise NotImplementedError("open_interest callback is not implemented")

    async def index(self, index: dict, symbol=None, **kwargs):
        """
        Unsupported, feed list https://docs.ccxt.com/en/latest/ccxt.pro.manual.html?rtd_search=fetchLedger#prerequisites
        :param index: the ccxt index dict
        :param symbol: the feed symbol
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported
        raise NotImplementedError("index callback is not implemented")

    async def orders(self, orders: list, **kwargs):
        """
        :param orders: the ccxt orders list
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported (ccxt is supporting it)
        raise NotImplementedError("orders callback is not implemented")
        symbol = None  # todo
        adapted = self.adapter.adapt_orders(orders, symbol=symbol)
        await self.push_to_channel(trading_constants.ORDERS_CHANNEL, adapted)

    async def trades(self, trades: list, **kwargs):
        """
        :param trades: the ccxt trades list
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported (ccxt is supporting it)
        raise NotImplementedError("trades callback is not implemented")
        adapted = self.adapter.adapt_trades(self.exchange.parse_trade(trades))
        await self.push_to_channel(trading_constants.TRADES_CHANNEL, adapted)

    async def balance(self, balance: dict, **kwargs):
        """
        :param balance: the ccxt balance dict
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported (ccxt is supporting it)
        raise NotImplementedError("balance callback is not implemented")
        await self.push_to_channel(
            trading_constants.BALANCE_CHANNEL,
            self.adapter.parse_balance(balance.balance),
        )

    async def transaction(self, transaction: dict, **kwargs):
        """
        :param transaction: the ccxt transaction dict
        :param kwargs: the feed kwargs
        """
        # TODO update this when supported (ccxt is supporting it). Use watchLedger ?
        raise NotImplementedError("transaction callback is not implemented")

    def _register_previous_open_candle(self, time_frame, symbol, candle):
        try:
            self._previous_open_candles[time_frame][symbol] = candle
        except KeyError:
            if time_frame not in self._previous_open_candles:
                self._previous_open_candles[time_frame] = {}
            self._previous_open_candles[time_frame][symbol] = candle

    def _get_previous_open_candle(self, time_frame, symbol):
        try:
            return self._previous_open_candles[time_frame][symbol]
        except KeyError:
            return None

    def _register_subsequent_unordered_candle(
        self, time_frame, symbol, parsed_timeframe, current_candle_time
    ):
        try:
            count, last_candle_time = self._subsequent_unordered_candles_count[
                time_frame
            ][symbol]
            if (
                current_candle_time - last_candle_time
                >= commons_enums.TimeFramesMinutes[parsed_timeframe]
                * commons_constants.MINUTE_TO_SECONDS
                * 1.5
            ):
                # candle is not subsequent from the previous one: reset count
                count = 0
            self._subsequent_unordered_candles_count[time_frame][symbol] = (
                count + 1,
                current_candle_time,
            )
            return False
        except KeyError:
            if time_frame not in self._subsequent_unordered_candles_count:
                self._subsequent_unordered_candles_count[time_frame] = {}
            self._subsequent_unordered_candles_count[time_frame][symbol] = (
                1,
                current_candle_time,
            )
            return True

    def _get_subsequent_unordered_candles_count(self, time_frame, symbol):
        try:
            return self._subsequent_unordered_candles_count[time_frame][symbol][0]
        except KeyError:
            return 0

    def _increment_error_counter(self, time_frame, error):
        error_key = error.__class__.__name__
        try:
            self._errors_count[time_frame][error_key] += 1
        except KeyError:
            if time_frame not in self._errors_count:
                self._errors_count[time_frame] = {}
            self._errors_count[time_frame][error_key] = 1
        return self._errors_count[time_frame][error_key]

    def _fix_candles_timestamps(self, candles: list, time_frame: commons_enums.TimeFrames) -> list:
        for candle in candles:
            candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] = self._get_utc_candle_timestamp(
                candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value], time_frame
            )
        return candles

    def _get_utc_candle_timestamp(self, timestamp: float, time_frame: commons_enums.TimeFrames) -> float:
        uniformized_timestamp = self.adapter.get_uniformized_timestamp(timestamp)
        if self.candle_delay_seconds != 0:
            return uniformized_timestamp + self.candle_delay_seconds
        current_time = self.exchange_manager.exchange.get_exchange_current_time()
        if current_time % commons_constants.MINUTE_TO_SECONDS > 20:
            # avoid new candle side effects: if current time is less than 20 seconds older than the current minute, ignore it
            expected_timestamp = time_frame_manager.get_last_timeframe_time(time_frame, current_time)
            delay = expected_timestamp - uniformized_timestamp
            hours_delay = delay / commons_constants.HOURS_TO_SECONDS
            if int(hours_delay) != hours_delay:
                # should not happen: raise if it does
                raise ValueError(
                    f"can't compute utc candle timestamp at this time (delay is not a whole number of hours, {delay=})"
                )
            self.candle_delay_seconds = delay
            self.logger.info(
                f"Applying websocket candle delay of {delay} seconds (= {hours_delay} hours)"
            )
            return uniformized_timestamp + self.candle_delay_seconds
        raise ValueError("can't compute utc candle timestamp at this time (too close to the next minute)")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest

import octobot_trading.enums as enums
import octobot_trading.exchange_data.contracts as contracts

from tests import event_loop

pytestmark = pytest.mark.asyncio

DEFAULT_SYMBOL = "BTC/USDT"


@pytest.mark.parametrize("contract_type,expected", [
    (enums.OptionContractType.INVERSE_EXPIRABLE, True),
    (enums.OptionContractType.LINEAR_EXPIRABLE, False),
])
async def test_is_inverse_contract(contract_type, expected):
    contract = contracts.OptionContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        contract_type=contract_type,
        current_leverage=decimal.Decimal(10))
    assert contract.is_inverse_contract() == expected

@pytest.mark.parametrize("position_mode,expected", [
    (enums.PositionMode.ONE_WAY, True),
    (enums.PositionMode.HEDGE, False),
])
async def test_is_one_way_position_mode(position_mode, expected):
    contract = contracts.OptionContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        contract_type=enums.OptionContractType.LINEAR_EXPIRABLE,
        position_mode=position_mode,
        current_leverage=decimal.Decimal(10))
    assert contract.is_one_way_position_mode() == expected


async def test_set_position_mode():
    contract = contracts.OptionContract(
        pair=DEFAULT_SYMBOL,
        margin_type=enums.MarginType.ISOLATED,
        contract_type=enums.OptionContractType.LINEAR_EXPIRABLE,
        current_leverage=decimal.Decimal(10))
    assert contract.position_mode == enums.PositionMode.ONE_WAY
    contract.set_position_mode(is_one_way=True)
    assert contract.position_mode == enums.PositionMode.ONE_WAY
    contract.set_position_mode(is_hedge=False)
    assert contract.position_mode == enums.PositionMode.ONE_WAY
    contract.set_position_mode(is_one_way=False)
    assert contract.position_mode == enums.PositionMode.HEDGE
    contract.set_position_mode(is_hedge=False)
    assert contract.position_mode == enums.PositionMode.ONE_WAY
    contract.set_position_mode(is_hedge=True)
    assert contract.position_mode == enums.PositionMode.HEDGE



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid

import octobot_commons.logging as logging


class Transaction:

    def __init__(self, exchange_name, creation_time, transaction_type, currency, symbol=None, transaction_id=None):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.transaction_id = transaction_id
        self.transaction_type = transaction_type
        self.creation_time = creation_time

        self.exchange_name = exchange_name
        self.symbol = symbol
        self.currency = currency

        # generate default transaction id
        if self.transaction_id is None:
            self.transaction_id = str(uuid.uuid4())

    def set_transaction_id(self, new_id):
        """
        Shouldn't be called outside of TransactionsManager.update_transaction_id
        to maintain TransactionsManager.transactions integrity
        :param new_id: the new transaction id
        """
        self.transaction_id = new_id



# pylint: disable=E0611, E0401
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.modes import abstract_trading_mode
from octobot_trading.modes.abstract_trading_mode import (
    AbstractTradingMode,
)

from octobot_trading.modes import channel
from octobot_trading.modes.channel import (
    check_factor,
    ModeChannelConsumer,
    ModeChannelProducer,
    ModeChannel,
    AbstractTradingModeConsumer,
    AbstractTradingModeProducer,
)

from octobot_trading.modes import scripted_trading_mode
from octobot_trading.modes.scripted_trading_mode import (
    AbstractScriptedTradingMode,
    AbstractScriptedTradingModeProducer,
)

from octobot_trading.modes import script_keywords
from octobot_trading.modes.script_keywords import (
    Context,
)

from octobot_trading.modes import mode_config
from octobot_trading.modes.mode_config import (
    get_activated_trading_mode,
    should_emit_trading_signals_user_input,
    is_trading_signal_emitter,
    user_select_order_amount,
    get_order_amount_value_desc,
    get_user_selected_order_amount,
)

from octobot_trading.modes import modes_factory
from octobot_trading.modes.modes_factory import (
    create_trading_modes,
    create_trading_mode,
    create_temporary_trading_mode_with_local_config,
)

from octobot_trading.modes import mode_activity
from octobot_trading.modes.mode_activity import (
    TradingModeActivity,
)

from octobot_trading.modes import modes_util
from octobot_trading.modes.modes_util import (
    get_required_candles_count,
    get_assets_requiring_extra_price_data_to_convert,
    convert_assets_to_target_asset,
    convert_asset_to_target_asset,
    get_instantly_filled_limit_order_adapted_price,
    get_instantly_filled_limit_order_adapted_price_and_quantity,
    notify_portfolio_optimization_complete,
    get_trading_modes_of_this_type_on_this_matrix,
    enabled_trader_only,
)

__all__ = [
    "ModeChannelConsumer",
    "ModeChannelProducer",
    "ModeChannel",
    "AbstractTradingModeProducer",
    "AbstractTradingMode",
    "AbstractTradingModeConsumer",
    "AbstractScriptedTradingMode",
    "AbstractScriptedTradingModeProducer",
    "Context",
    "check_factor",
    "create_trading_modes",
    "create_trading_mode",
    "create_temporary_trading_mode_with_local_config",
    "TradingModeActivity",
    "get_activated_trading_mode",
    "should_emit_trading_signals_user_input",
    "is_trading_signal_emitter",
    "user_select_order_amount",
    "get_order_amount_value_desc",
    "get_user_selected_order_amount",
    "get_required_candles_count",
    "get_assets_requiring_extra_price_data_to_convert",
    "convert_assets_to_target_asset",
    "convert_asset_to_target_asset",
    "get_instantly_filled_limit_order_adapted_price",
    "get_instantly_filled_limit_order_adapted_price_and_quantity",
    "notify_portfolio_optimization_complete",
    "get_trading_modes_of_this_type_on_this_matrix",
    "enabled_trader_only",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.constants as constants
import octobot_trading.exchange_data.prices.channel.price as prices_channel
import octobot_trading.exchange_data.prices.prices_manager as prices_manager
import octobot_trading.enums as enums


class MarkPriceUpdater(prices_channel.MarkPriceProducer):
    CHANNEL_NAME = constants.MARK_PRICE_CHANNEL

    MARK_PRICE_REFRESH_TIME = 7

    def __init__(self, channel):
        super().__init__(channel)
        self.recent_trades_consumer = None
        self.ticker_consumer = None
        self.refresh_time = MarkPriceUpdater.MARK_PRICE_REFRESH_TIME

    async def start(self):
        refresh_threshold = self.channel.exchange_manager.get_rest_pairs_refresh_threshold()
        if refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.MEDIUM:
            self.refresh_time = 12
        elif refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.SLOW:
            self.refresh_time = 17
        if self.channel.is_paused:
            await self.pause()
        else:
            if self._should_subscribe():
                await self.subscribe()
            elif self._should_fetch_on_exchange():
                await self.start_fetching()
            else:
                self.logger.debug(f"{self.__class__.__name__} wont be used, stopping...")

    async def subscribe(self):
        self.recent_trades_consumer = await exchanges_channel.get_chan(constants.RECENT_TRADES_CHANNEL,
                                                                       self.channel.exchange_manager.id) \
            .new_consumer(self.handle_recent_trades_update)
        self.ticker_consumer = await exchanges_channel.get_chan(constants.TICKER_CHANNEL,
                                                                self.channel.exchange_manager.id) \
            .new_consumer(self.handle_ticker_update)

    async def unsubscribe(self):
        if self.recent_trades_consumer:
            await exchanges_channel.get_chan(constants.RECENT_TRADES_CHANNEL, self.channel.exchange_manager.id) \
                .remove_consumer(self.recent_trades_consumer)
        if self.ticker_consumer:
            await exchanges_channel.get_chan(constants.TICKER_CHANNEL, self.channel.exchange_manager.id) \
                .remove_consumer(self.ticker_consumer)

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            await self.run()

    async def pause(self) -> None:
        await super().pause()
        if not self.channel.exchange_manager.is_future:
            await self.unsubscribe()

    async def handle_recent_trades_update(self, exchange: str, exchange_id: str,
                                          cryptocurrency: str, symbol: str, recent_trades: list):
        """
        Recent trades channel consumer callback
        """
        try:
            mark_price = prices_manager.calculate_mark_price_from_recent_trade_prices(
                [
                    decimal.Decimal(str(last_price[enums.ExchangeConstantsOrderColumns.PRICE.value]))
                    for last_price in recent_trades
                ]
            )

            await self.push(symbol, mark_price, mark_price_source=enums.MarkPriceSources.RECENT_TRADE_AVERAGE.value)
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle recent trades update : {e}")

    async def handle_ticker_update(self, exchange: str, exchange_id: str,
                                   cryptocurrency: str, symbol: str, ticker: dict):
        """
        Ticker channel consumer callback
        """
        try:
            if ticker[enums.ExchangeConstantsTickersColumns.CLOSE.value]:
                await self.push(symbol, decimal.Decimal(str(ticker[enums.ExchangeConstantsTickersColumns.CLOSE.value])),
                                mark_price_source=enums.MarkPriceSources.TICKER_CLOSE_PRICE.value)
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle ticker update : {e}")

    async def start_fetching(self):
        """
        Mark price updater from exchange data
        """
        while not self.should_stop and not self.channel.is_paused:
            try:
                for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    await self.fetch_market_price(pair)
            except (errors.NotSupported, NotImplementedError):
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            finally:
                await asyncio.sleep(self.refresh_time)

    async def fetch_market_price(self, symbol: str):
        try:
            if self.channel.exchange_manager.exchange.FUNDING_WITH_MARK_PRICE:
                mark_price, funding_rate = await self.channel.exchange_manager.exchange. \
                    get_mark_price_and_funding(symbol)
                mark_price = decimal.Decimal(str(mark_price))
                await self.push_funding_rate(symbol, funding_rate)
            else:
                mark_price = await self.channel.exchange_manager.exchange.get_mark_price(symbol)

            if mark_price:
                await self.push(symbol, mark_price[enums.ExchangeConstantsMarkPriceColumns.MARK_PRICE.value])
        except (errors.NotSupported, NotImplementedError) as ne:
            raise ne
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update funding rate : {html_util.get_html_summary_if_relevant(e)}")
        return None

    async def push_funding_rate(self, symbol, funding_rate):
        if funding_rate:
            predicted_funding_rate = \
                funding_rate.get(enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value, constants.NaN)
            await exchanges_channel.get_chan(constants.FUNDING_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                push(symbol,
                     decimal.Decimal(str(funding_rate[enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value])),
                     decimal.Decimal(str(predicted_funding_rate or constants.NaN)),
                     funding_rate[enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value],
                     funding_rate[enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value])

    def _should_fetch_on_exchange(self) -> bool:
        return not (
            self.channel.exchange_manager.exchange.FUNDING_WITH_MARK_PRICE
            or self.channel.exchange_manager.exchange.MARK_PRICE_IN_TICKER
            or self.channel.exchange_manager.exchange.MARK_PRICE_IN_POSITION
        )

    def _should_subscribe(self):
        return not self.channel.exchange_manager.is_future or \
            not self.channel.exchange_manager.exchange.MARK_PRICE_IN_POSITION



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import octobot_trading.enums as enums
import typing


async def inner_test_initialize_without_kwargs(order, status: enums.OrderStatus, mocked_method_name: typing.Optional[str]):
    order.status = status
    order.exchange_manager.is_backtesting = True
    pre_call_states = []
    if mocked_method_name is not None:
        async def _side_effect(*args, **kwargs):
            pre_call_states.append(order.state)
            return await origin_method(*args, **kwargs)

        origin_method = getattr(order, mocked_method_name)
        with mock.patch.object(order, mocked_method_name, mock.AsyncMock(side_effect=_side_effect)) as method_mock:
            await order.initialize()
            method_mock.assert_awaited_once()
    else:
        await order.initialize()
        pre_call_states.append(order.state)
    pre_call_state = pre_call_states[0]
    assert pre_call_state.is_from_exchange_data is False
    assert pre_call_state.enable_associated_orders_creation is True
    assert pre_call_state.is_already_counted_in_available_funds is False
    order.clear()


async def inner_test_initialize_with_kwargs(
    order, status: enums.OrderStatus, 
    mocked_method_name: typing.Optional[str], 
    expected_is_from_exchange_data: bool = True
):
    order.status = status
    order.exchange_manager.is_backtesting = True
    pre_call_states = []
    if mocked_method_name is not None:
        async def _side_effect(*args, **kwargs):
            pre_call_states.append(order.state)
            return await origin_method(*args, **kwargs)

        origin_method = getattr(order, mocked_method_name)
        with mock.patch.object(order, mocked_method_name, mock.AsyncMock(side_effect=_side_effect)) as method_mock:
            await order.initialize(
                is_from_exchange_data=True, enable_associated_orders_creation=False, 
                is_already_counted_in_available_funds=True
            )
            method_mock.assert_awaited_once()
    else:
        await order.initialize(
            is_from_exchange_data=True, enable_associated_orders_creation=False, 
            is_already_counted_in_available_funds=True
        )
        pre_call_states.append(order.state)
    pre_call_state = pre_call_states[0]
    assert pre_call_state.is_from_exchange_data is expected_is_from_exchange_data
    assert pre_call_state.enable_associated_orders_creation is False
    assert pre_call_state.is_already_counted_in_available_funds is True
    order.clear()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import decimal
import octobot_commons.dataclasses

import octobot_trading.personal_data.orders.trailing_profiles.trailing_profile_types as trailing_profile_types

@dataclasses.dataclass
class TrailingProfile(octobot_commons.dataclasses.FlexibleDataclass):

    @staticmethod
    def get_type() -> trailing_profile_types.TrailingProfileTypes:
        raise NotImplementedError("get_name is not implemented")

    def update_and_get_trailing_price(self, updated_price: decimal.Decimal) -> decimal.Decimal:
        raise NotImplementedError("update_and_get_trailing_price is not implemented")

    def to_dict(self) -> dict:
        return dataclasses.asdict(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.portfolios import channel
from octobot_trading.personal_data.portfolios.channel import (
    BalanceUpdater,
    BalanceProfitabilityUpdater,
    BalanceUpdaterSimulator,
    BalanceProfitabilityUpdaterSimulator,
    BalanceProducer,
    BalanceChannel,
    BalanceProfitabilityProducer,
    BalanceProfitabilityChannel,
)

from octobot_trading.personal_data.portfolios import portfolio
from octobot_trading.personal_data.portfolios.portfolio import (
    Portfolio,
)
from octobot_trading.personal_data.portfolios import asset
from octobot_trading.personal_data.portfolios.asset import (
    Asset,
)
from octobot_trading.personal_data.portfolios import portfolio_factory
from octobot_trading.personal_data.portfolios import portfolio_profitability
from octobot_trading.personal_data.portfolios import sub_portfolio
from octobot_trading.personal_data.portfolios import portfolio_manager
from octobot_trading.personal_data.portfolios import portfolio_value_holder
from octobot_trading.personal_data.portfolios import portfolio_value_holder_factory
from octobot_trading.personal_data.portfolios import value_converter
from octobot_trading.personal_data.portfolios import types
from octobot_trading.personal_data.portfolios import portfolio_util
from octobot_trading.personal_data.portfolios import history
from octobot_trading.personal_data.portfolios import sub_portfolio_data
from octobot_trading.personal_data.portfolios import resolved_orders_portfolio_delta

from octobot_trading.personal_data.portfolios.portfolio_factory import (
    create_portfolio_from_exchange_manager,
)
from octobot_trading.personal_data.portfolios.portfolio_profitability import (
    PortfolioProfitability,
)
from octobot_trading.personal_data.portfolios.sub_portfolio import (
    SubPortfolio,
)
from octobot_trading.personal_data.portfolios.portfolio_manager import (
    PortfolioManager,
)
from octobot_trading.personal_data.portfolios.value_converter import (
    ValueConverter,
)
from octobot_trading.personal_data.portfolios.portfolio_value_holder import (
    PortfolioValueHolder
)
from octobot_trading.personal_data.portfolios.holders import (
    FuturesPortfolioValueHolder,
    OptionPortfolioValueHolder,
    MarginPortfolioValueHolder,
    SpotPortfolioValueHolder,
)
from octobot_trading.personal_data.portfolios.portfolio_value_holder_factory import (
    create_portfolio_value_holder,
)
from octobot_trading.personal_data.portfolios.types import (
    FuturePortfolio,
    OptionPortfolio,
    MarginPortfolio,
    SpotPortfolio,
)
from octobot_trading.personal_data.portfolios.assets import (
    FutureAsset,
    OptionAsset,
    MarginAsset,
    SpotAsset,
)
from octobot_trading.personal_data.portfolios.portfolio_util import (
    parse_decimal_portfolio,
    parse_decimal_config_portfolio,
    format_dict_portfolio_values,
    filter_empty_values,
    portfolio_to_float,
    get_draw_down,
    get_coefficient_of_determination,
    get_asset_price_from_converter_or_tickers,
    resolve_sub_portfolios,
    get_portfolio_filled_orders_deltas,
    get_accepted_missed_deltas,
    get_master_checked_sub_portfolio_update,
    get_assets_delta_from_orders,
    get_fees_only_asset_deltas_from_orders,
)
from octobot_trading.personal_data.portfolios.history import (
    create_historical_asset_value_from_dict_like_object,
    HistoricalAssetValue,
    HistoricalPortfolioValueManager,
)
from octobot_trading.personal_data.portfolios.sub_portfolio_data import (
    SubPortfolioData,
)
from octobot_trading.personal_data.portfolios.resolved_orders_portfolio_delta import (
    ResolvedOrdersPortoflioDelta,
)
    
__all__ = [
    "BalanceUpdaterSimulator",
    "BalanceProfitabilityUpdaterSimulator",
    "create_portfolio_from_exchange_manager",
    "BalanceUpdater",
    "BalanceProfitabilityUpdater",
    "PortfolioProfitability",
    "Portfolio",
    "Asset",
    "BalanceProducer",
    "BalanceChannel",
    "BalanceProfitabilityProducer",
    "BalanceProfitabilityChannel",
    "SubPortfolio",
    "PortfolioManager",
    "ValueConverter",
    "PortfolioValueHolder",
    "FuturesPortfolioValueHolder",
    "OptionPortfolioValueHolder",
    "MarginPortfolioValueHolder",
    "SpotPortfolioValueHolder",
    "create_portfolio_value_holder",
    "FuturePortfolio",
    "OptionPortfolio",
    "MarginPortfolio",
    "SpotPortfolio",
    "FutureAsset",
    "OptionAsset",
    "MarginAsset",
    "SpotAsset",
    "parse_decimal_portfolio",
    "parse_decimal_config_portfolio",
    "format_dict_portfolio_values",
    "filter_empty_values",
    "portfolio_to_float",
    "get_draw_down",
    "get_coefficient_of_determination",
    "get_asset_price_from_converter_or_tickers",
    "resolve_sub_portfolios",
    "get_portfolio_filled_orders_deltas",
    "get_accepted_missed_deltas",
    "get_master_checked_sub_portfolio_update",
    "get_assets_delta_from_orders",
    "get_fees_only_asset_deltas_from_orders",
    "create_historical_asset_value_from_dict_like_object",
    "get_draw_down",
    "HistoricalAssetValue",
    "HistoricalPortfolioValueManager",
    "SubPortfolioData",
    "ResolvedOrdersPortoflioDelta",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data.contracts import contract
from octobot_trading.exchange_data.contracts.contract import (
    Contract,
)

from octobot_trading.exchange_data.contracts import margin_contract
from octobot_trading.exchange_data.contracts.margin_contract import (
    MarginContract,
)

from octobot_trading.exchange_data.contracts import future_contract
from octobot_trading.exchange_data.contracts.future_contract import (
    FutureContract,
)

from octobot_trading.exchange_data.contracts import option_contract
from octobot_trading.exchange_data.contracts.option_contract import (
    OptionContract,
)

from octobot_trading.exchange_data.contracts import contract_factory
from octobot_trading.exchange_data.contracts.contract_factory import (
    get_contract_type_from_symbol,
    update_contracts_from_positions,
    update_future_contract_from_dict,
    create_default_future_contract,
    create_default_option_contract,
    create_contract,
)

__all__ = [
    "Contract",
    "MarginContract",
    "FutureContract",
    "OptionContract",
    "get_contract_type_from_symbol",
    "update_contracts_from_positions",
    "update_future_contract_from_dict",
    "create_default_future_contract",
    "create_default_option_contract",
    "create_contract",
]
